#
# This file is part of pysmi software.
#
# Copyright (c) 2015-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysmi/license.html
#
import os
import sys

import ply.yacc as yacc
from pysmi import debug
from pysmi import error
from pysmi.lexer.smi import lexerFactory
from pysmi.parser.base import AbstractParser

YACC_VERSION = [int(x) for x in yacc.__version__.split(".")]


# noinspection PyMethodMayBeStatic,PyIncorrectDocstring
class SmiV2Parser(AbstractParser):
    defaultLexer = lexerFactory()

    def __init__(self, startSym="mibFile", tempdir=""):
        if tempdir:
            tempdir = os.path.join(tempdir, startSym)
            try:
                os.makedirs(tempdir)
            except OSError:
                if sys.exc_info()[1].errno != 17:
                    raise error.PySmiError(
                        f"Failed to create cache directory {tempdir}: {sys.exc_info()[1]}"
                    )

        self.lexer = self.defaultLexer(tempdir=tempdir)

        # tokens are required for parser
        self.tokens = self.lexer.tokens

        if YACC_VERSION < [3, 0]:
            self.parser = yacc.yacc(
                module=self,
                start=startSym,
                write_tables=bool(tempdir),
                debug=False,
                outputdir=tempdir,
            )
        else:
            if debug.logger & debug.FLAG_PARSER:
                logger = debug.logger.get_current_logger()
            else:
                logger = yacc.NullLogger()

            if debug.logger & debug.FLAG_GRAMMAR:
                debuglogger = debug.logger.get_current_logger()
            else:
                debuglogger = None

            self.parser = yacc.yacc(
                module=self,
                start=startSym,
                write_tables=bool(tempdir),
                debug=False,
                outputdir=tempdir,
                debuglog=debuglogger,
                errorlog=logger,
            )

    def reset(self):
        # Ply requires lexer reinitialization for (at least) resetting lineno
        self.lexer.reset()

    def parse(self, data, **kwargs):
        debug.logger & debug.FLAG_PARSER and debug.logger(
            f'source MIB size is {len(data)} characters, first 50 characters are "{data[:50]}..."'
        )

        ast = self.parser.parse(data, lexer=self.lexer.lexer)

        self.reset()

        if ast and ast[0] == "mibFile" and ast[1]:  # mibfile is not empty
            return ast[1]
        else:
            return []

    #
    # SMIv2 grammar follows
    #

    def p_mibFile(self, p):
        """mibFile : modules
        | empty"""
        p[0] = ("mibFile", p[1])

    def p_modules(self, p):
        """modules : modules module
        | module"""
        n = len(p)
        if n == 3:
            p[0] = p[1] + [p[2]]
        elif n == 2:
            p[0] = [p[1]]

    def p_module(self, p):
        """module : moduleName moduleOid DEFINITIONS COLON_COLON_EQUAL BEGIN exportsClause linkagePart declarationPart END"""
        p[0] = (
            p[1],  # name
            p[2],  # oid
            p[7] or {},  # linkage (imports)
            p[8],
        )  # declaration

    def p_moduleOid(self, p):
        """moduleOid : '{' objectIdentifier '}'
        | empty"""
        n = len(p)
        if n == 4:
            p[0] = p[2]

    def p_linkagePart(self, p):
        """linkagePart : linkageClause
        | empty"""
        if p[1]:
            p[0] = p[1]

    def p_linkageClause(self, p):
        """linkageClause : IMPORTS importPart ';'"""
        p[0] = p[2]

    def p_exportsClause(self, p):
        """exportsClause : EXPORTS
        | empty"""

    def p_importPart(self, p):
        """importPart : imports
        | empty"""
        # libsmi: TODO: ``IMPORTS ;'' allowed? refer ASN.1!
        if p[1]:
            importDict = {}
            for imp in p[1]:  # don't do just dict() because moduleNames may be repeated
                fromModule, symbols = imp
                if fromModule in importDict:
                    importDict[fromModule] += symbols
                else:
                    importDict[fromModule] = symbols

            p[0] = importDict

    def p_imports(self, p):
        """imports : imports import
        | import"""
        n = len(p)
        if n == 3:
            p[0] = p[1] + [p[2]]
        elif n == 2:
            p[0] = [p[1]]

    def p_import(self, p):
        """import : importIdentifiers FROM moduleName"""
        # libsmi: TODO: multiple clauses with same moduleName allowed?
        # I guess so. refer ASN.1!
        p[0] = (p[3], p[1])  # moduleName  # ids

    def p_importIdentifiers(self, p):
        """importIdentifiers : importIdentifiers ',' importIdentifier
        | importIdentifier"""
        n = len(p)
        if n == 4:
            p[0] = p[1] + [p[3]]
        elif n == 2:
            p[0] = [p[1]]

    # Note that some named types must not be imported, REF:RFC1902,590
    def p_importIdentifier(self, p):
        """importIdentifier : LOWERCASE_IDENTIFIER
        | UPPERCASE_IDENTIFIER
        | importedKeyword"""
        p[0] = p[1]

    def p_importedKeyword(self, p):
        """importedKeyword : importedSMIKeyword
        | BITS
        | INTEGER32
        | IPADDRESS
        | MANDATORY_GROUPS
        | MODULE_COMPLIANCE
        | MODULE_IDENTITY
        | OBJECT_GROUP
        | OBJECT_IDENTITY
        | OBJECT_TYPE
        | OPAQUE
        | TEXTUAL_CONVENTION
        | TIMETICKS
        | UNSIGNED32"""
        p[0] = p[1]

    def p_importedSMIKeyword(self, p):
        """importedSMIKeyword : AGENT_CAPABILITIES
        | COUNTER32
        | COUNTER64
        | GAUGE32
        | NOTIFICATION_GROUP
        | NOTIFICATION_TYPE
        | TRAP_TYPE"""
        p[0] = p[1]

    def p_moduleName(self, p):
        """moduleName : UPPERCASE_IDENTIFIER"""
        p[0] = p[1]

    def p_declarationPart(self, p):
        """declarationPart : declarations
        | empty"""
        if p[1]:
            p[0] = p[1]

    def p_declarations(self, p):
        """declarations : declarations declaration
        | declaration"""
        n = len(p)
        if n == 3:
            p[0] = p[1] + [p[2]]
        elif n == 2:
            p[0] = [p[1]]

    def p_declaration(self, p):
        """declaration : typeDeclaration
        | valueDeclaration
        | objectIdentityClause
        | objectTypeClause
        | trapTypeClause
        | notificationTypeClause
        | moduleIdentityClause
        | moduleComplianceClause
        | objectGroupClause
        | notificationGroupClause
        | agentCapabilitiesClause
        | macroClause"""
        if p[1]:
            p[0] = p[1]

    def p_macroClause(self, p):
        """macroClause : macroName MACRO END"""

    def p_macroName(self, p):
        """macroName : MODULE_IDENTITY
        | OBJECT_TYPE
        | TRAP_TYPE
        | NOTIFICATION_TYPE
        | OBJECT_IDENTITY
        | TEXTUAL_CONVENTION
        | OBJECT_GROUP
        | NOTIFICATION_GROUP
        | MODULE_COMPLIANCE
        | AGENT_CAPABILITIES"""

    def p_choiceClause(self, p):
        """choiceClause : CHOICE"""

    # libsmi: The only ASN.1 value declarations are for OIDs, REF:RFC1902,491.
    def p_fuzzy_lowercase_identifier(self, p):
        """fuzzy_lowercase_identifier : LOWERCASE_IDENTIFIER
        | UPPERCASE_IDENTIFIER"""
        p[0] = p[1]

    def p_valueDeclaration(self, p):
        """valueDeclaration : fuzzy_lowercase_identifier OBJECT IDENTIFIER COLON_COLON_EQUAL '{' objectIdentifier '}'"""
        p[0] = ("valueDeclaration", p[1], p[6])  # id  # objectIdentifier

    def p_typeDeclaration(self, p):
        """typeDeclaration : typeName COLON_COLON_EQUAL typeDeclarationRHS"""
        p[0] = ("typeDeclaration", p[1], p[3])  # name  # declarationRHS

    def p_typeName(self, p):
        """typeName : UPPERCASE_IDENTIFIER
        | typeSMI"""
        p[0] = p[1]

    def p_typeSMI(self, p):
        """typeSMI : typeSMIandSPPI
        | typeSMIonly"""
        p[0] = p[1]

    def p_typeSMIandSPPI(self, p):
        """typeSMIandSPPI : IPADDRESS
        | TIMETICKS
        | OPAQUE
        | INTEGER32
        | UNSIGNED32"""
        p[0] = p[1]

    def p_typeSMIonly(self, p):
        """typeSMIonly : COUNTER32
        | GAUGE32
        | COUNTER64"""
        p[0] = p[1]

    def p_typeDeclarationRHS(self, p):
        """typeDeclarationRHS : Syntax
        | TEXTUAL_CONVENTION DisplayPart STATUS Status DESCRIPTION Text ReferPart SYNTAX Syntax
        | choiceClause"""
        if p[1]:
            if p[1] == "TEXTUAL-CONVENTION":
                p[0] = (
                    "typeDeclarationRHS",
                    p[2],  # display
                    p[4],  # status
                    (p[5], p[6]),  # description
                    p[7],  # reference
                    p[9],
                )  # syntax
            else:
                p[0] = ("typeDeclarationRHS", p[1])
                # ignore the choiceClause

    def p_conceptualTable(self, p):
        """conceptualTable : SEQUENCE OF row"""
        p[0] = ("conceptualTable", p[3])

    def p_row(self, p):
        """row : UPPERCASE_IDENTIFIER"""
        # libsmi: TODO: this must be an entryType
        p[0] = ("row", p[1])

    def p_entryType(self, p):
        """entryType : SEQUENCE '{' sequenceItems '}'"""
        p[0] = (p[1], p[3])

    def p_sequenceItems(self, p):
        """sequenceItems : sequenceItems ',' sequenceItem
        | sequenceItem"""
        # libsmi: TODO: might this list be emtpy?
        n = len(p)
        if n == 4:
            p[0] = p[1] + [p[3]]
        elif n == 2:
            p[0] = [p[1]]

    def p_sequenceItem(self, p):
        """sequenceItem : LOWERCASE_IDENTIFIER sequenceSyntax"""
        p[0] = (p[1], p[2])

    def p_Syntax(self, p):
        """Syntax : ObjectSyntax
        | BITS '{' NamedBits '}'"""
        # libsmi: TODO: standalone `BITS' ok? seen in RMON2-MIB
        # libsmi: -> no, it's only allowed in a SEQUENCE {...}
        n = len(p)
        if n == 2:
            p[0] = p[1]
        elif n == 5:
            p[0] = (p[1], p[3])

    def p_sequenceSyntax(self, p):
        """sequenceSyntax : BITS
        | UPPERCASE_IDENTIFIER anySubType
        | sequenceObjectSyntax"""
        p[0] = p[1]  # no subtype or complex syntax supported

    def p_NamedBits(self, p):
        """NamedBits : NamedBits ',' NamedBit
        | NamedBit"""
        n = len(p)
        if n == 4:
            p[0] = p[1] + [p[3]]
        elif n == 2:
            p[0] = [p[1]]

    def p_NamedBit(self, p):
        """NamedBit : LOWERCASE_IDENTIFIER '(' NUMBER ')'"""
        p[0] = (p[1], p[3])

    def p_objectIdentityClause(self, p):
        """objectIdentityClause : LOWERCASE_IDENTIFIER OBJECT_IDENTITY STATUS Status DESCRIPTION Text ReferPart COLON_COLON_EQUAL '{' objectIdentifier '}'"""
        p[0] = (
            "objectIdentityClause",
            p[1],  # id
            #  p[2], # OBJECT_IDENTITY
            p[4],  # status
            (p[5], p[6]),  # description
            p[7],  # reference
            p[10],
        )  # objectIdentifier

    def p_objectTypeClause(self, p):
        """objectTypeClause : LOWERCASE_IDENTIFIER OBJECT_TYPE SYNTAX Syntax UnitsPart MaxOrPIBAccessPart STATUS Status descriptionClause ReferPart IndexPart MibIndex DefValPart COLON_COLON_EQUAL '{' ObjectName '}'"""
        p[0] = (
            "objectTypeClause",
            p[1],  # id
            #  p[2], # OBJECT_TYPE
            p[4],  # syntax
            p[5],  # UnitsPart
            p[6],  # MaxOrPIBAccessPart
            p[8],  # status
            p[9],  # descriptionClause
            p[10],  # reference
            p[11],  # augmentions
            p[12],  # index
            p[13],  # DefValPart
            p[16],
        )  # ObjectName

    def p_descriptionClause(self, p):
        """descriptionClause : DESCRIPTION Text
        | empty"""
        if p[1]:
            p[0] = (p[1], p[2])

    def p_trapTypeClause(self, p):
        """trapTypeClause : fuzzy_lowercase_identifier TRAP_TYPE ENTERPRISE objectIdentifier VarPart DescrPart ReferPart COLON_COLON_EQUAL NUMBER"""
        # libsmi: TODO: range of number?
        p[0] = (
            "trapTypeClause",
            p[1],  # fuzzy_lowercase_identifier
            #  p[2], # TRAP_TYPE
            p[4],  # objectIdentifier
            p[5],  # VarPart
            p[6],  # description
            p[7],  # reference
            p[9],
        )  # NUMBER

    def p_VarPart(self, p):
        """VarPart : VARIABLES '{' VarTypes '}'
        | empty"""
        p[0] = p[1] and p[3] or []

    def p_VarTypes(self, p):
        """VarTypes : VarTypes ',' VarType
        | VarType"""
        n = len(p)
        if n == 4:
            p[0] = ("VarTypes", p[1][1] + [p[3]])
        elif n == 2:
            p[0] = ("VarTypes", [p[1]])

    def p_VarType(self, p):
        """VarType : ObjectName"""
        p[0] = p[1][1][0]

    def p_DescrPart(self, p):
        """DescrPart : DESCRIPTION Text
        | empty"""
        if p[1]:
            p[0] = (p[1], p[2])

    def p_MaxOrPIBAccessPart(self, p):
        """MaxOrPIBAccessPart : MaxAccessPart
        | empty"""
        if p[1]:
            p[0] = p[1]

    def p_MaxAccessPart(self, p):
        """MaxAccessPart : MAX_ACCESS Access
        | ACCESS Access"""
        p[0] = ("MaxAccessPart", p[2])

    def p_notificationTypeClause(self, p):
        """notificationTypeClause : LOWERCASE_IDENTIFIER NOTIFICATION_TYPE NotificationObjectsPart STATUS Status DESCRIPTION Text ReferPart COLON_COLON_EQUAL '{' NotificationName '}'"""
        p[0] = (
            "notificationTypeClause",
            p[1],  # id
            #  p[2], # NOTIFICATION_TYPE
            p[3],  # NotificationObjectsPart
            p[5],  # status
            (p[6], p[7]),  # description
            p[8],  # reference
            p[11],
        )  # NotificationName aka objectIdentifier

    def p_moduleIdentityClause(self, p):
        """moduleIdentityClause : LOWERCASE_IDENTIFIER MODULE_IDENTITY SubjectCategoriesPart LAST_UPDATED ExtUTCTime ORGANIZATION Text CONTACT_INFO Text DESCRIPTION Text RevisionPart COLON_COLON_EQUAL '{' objectIdentifier '}'"""
        p[0] = (
            "moduleIdentityClause",
            p[1],  # id
            #  p[2], # MODULE_IDENTITY
            # XXX  p[3], # SubjectCategoriesPart
            (p[4], p[5]),  # last updated
            (p[6], p[7]),  # organization
            (p[8], p[9]),  # contact info
            (p[10], p[11]),  # description
            p[12],  # RevisionPart
            p[15],
        )  # objectIdentifier

    # Subject categories: RFC3159

    def p_SubjectCategoriesPart(self, p):
        """SubjectCategoriesPart : SUBJECT_CATEGORIES '{' SubjectCategories '}'
        | empty"""
        # if p[1]:
        #  p[0] = (p[1], p[3])

    def p_SubjectCategories(self, p):
        """SubjectCategories : CategoryIDs"""
        # p[0] = p[1]

    def p_CategoryIDs(self, p):
        """CategoryIDs : CategoryIDs ',' CategoryID
        | CategoryID"""
        # n = len(p)
        # if n == 4:
        #  p[0] = ('CategoryIDs', p[1][1] + [p[3]])
        # elif n == 2:
        #  p[0] = ('CategoryIDs', [p[1]])

    def p_CategoryID(self, p):
        """CategoryID : LOWERCASE_IDENTIFIER '(' NUMBER ')'
        | LOWERCASE_IDENTIFIER"""
        # n = len(p)
        # if n == 2:
        #  p[0] = ('CategoryID', p[1])
        # elif n == 5:
        #  p[0] = ('CategoryID', p[3])

    # ...subject categories

    def p_ObjectSyntax(self, p):
        """ObjectSyntax : SimpleSyntax
        | conceptualTable
        | row
        | entryType
        | ApplicationSyntax
        | typeTag SimpleSyntax"""
        n = len(p)
        if n == 2:
            p[0] = p[1]
        elif n == 3:
            p[0] = p[2]

    def p_typeTag(self, p):
        """typeTag : '[' APPLICATION NUMBER ']' IMPLICIT
        | '[' UNIVERSAL NUMBER ']' IMPLICIT"""

    def p_sequenceObjectSyntax(self, p):
        """sequenceObjectSyntax : sequenceSimpleSyntax
        | sequenceApplicationSyntax"""
        # libsmi: TO DO: add to this rule conceptualTable, row, entryType
        p[0] = p[1]

    def p_valueofObjectSyntax(self, p):
        """valueofObjectSyntax : valueofSimpleSyntax"""
        p[0] = p[1]

    def p_SimpleSyntax(self, p):
        """SimpleSyntax : INTEGER
        | INTEGER integerSubType
        | INTEGER enumSpec
        | INTEGER32
        | INTEGER32 integerSubType
        | UPPERCASE_IDENTIFIER enumSpec
        | UPPERCASE_IDENTIFIER integerSubType
        | OCTET STRING
        | OCTET STRING octetStringSubType
        | UPPERCASE_IDENTIFIER octetStringSubType
        | OBJECT IDENTIFIER anySubType"""
        n = len(p)
        if n == 2:
            p[0] = ("SimpleSyntax", p[1])

        elif n == 3:
            if p[1] == "OCTET":
                p[0] = ("SimpleSyntax", p[1] + " " + p[2])
            else:
                p[0] = ("SimpleSyntax", p[1], p[2])

        elif n == 4:
            p[0] = ("SimpleSyntax", p[1] + " " + p[2], p[3])

    def p_valueofSimpleSyntax(self, p):
        """valueofSimpleSyntax : NUMBER
        | NEGATIVENUMBER
        | NUMBER64
        | NEGATIVENUMBER64
        | HEX_STRING
        | BIN_STRING
        | LOWERCASE_IDENTIFIER
        | QUOTED_STRING
        | '{' objectIdentifier_defval '}'"""
        # libsmi for objectIdentifier_defval:
        # This is only for some MIBs with invalid numerical
        # OID notation for DEFVALs. We DO NOT parse them
        # correctly. We just don't want to produce a
        # parser error.
        n = len(p)
        if n == 2:
            p[0] = p[1]
        elif n == 4:  # XXX
            pass

    def p_sequenceSimpleSyntax(self, p):
        """sequenceSimpleSyntax : INTEGER anySubType
        | INTEGER32 anySubType
        | OCTET STRING anySubType
        | OBJECT IDENTIFIER anySubType"""
        n = len(p)
        if n == 3:
            p[0] = p[1]  # XXX not supporting subtypes here
        elif n == 4:
            p[0] = p[1] + " " + p[2]  # XXX not supporting subtypes here

    def p_ApplicationSyntax(self, p):
        """ApplicationSyntax : IPADDRESS anySubType
        | COUNTER32
        | COUNTER32 integerSubType
        | GAUGE32
        | GAUGE32 integerSubType
        | UNSIGNED32
        | UNSIGNED32 integerSubType
        | TIMETICKS anySubType
        | OPAQUE
        | OPAQUE octetStringSubType
        | COUNTER64
        | COUNTER64 integerSubType"""
        # COUNTER32 and COUNTER64 was with anySubType in libsmi
        n = len(p)
        if n == 2:
            p[0] = ("ApplicationSyntax", p[1])
        elif n == 3:
            p[0] = ("ApplicationSyntax", p[1], p[2])

    def p_sequenceApplicationSyntax(self, p):
        """sequenceApplicationSyntax : IPADDRESS anySubType
        | COUNTER32 anySubType
        | GAUGE32 anySubType
        | UNSIGNED32 anySubType
        | TIMETICKS anySubType
        | OPAQUE
        | COUNTER64 anySubType"""
        n = len(p)
        if n == 2:
            p[0] = p[1]
        elif n == 3:
            p[0] = p[1]  # XXX not supporting subtypes here

    def p_anySubType(self, p):
        """anySubType : integerSubType
        | octetStringSubType
        | enumSpec
        | empty"""
        if p[1]:
            p[0] = p[1]

    def p_integerSubType(self, p):
        """integerSubType : '(' ranges ')'"""
        p[0] = ("integerSubType", p[2])

    def p_octetStringSubType(self, p):
        """octetStringSubType : '(' SIZE '(' ranges ')' ')'"""
        p[0] = ("octetStringSubType", p[4])

    def p_ranges(self, p):
        """ranges : ranges '|' range
        | range"""
        n = len(p)
        if n == 4:
            p[0] = p[1] + [p[3]]
        elif n == 2:
            p[0] = [p[1]]

    def p_range(self, p):
        """range : value DOT_DOT value
        | value"""
        n = len(p)
        if n == 2:
            p[0] = (p[1],)
        elif n == 4:
            p[0] = (p[1], p[3])

    def p_value(self, p):
        """value : NEGATIVENUMBER
        | NUMBER
        | NEGATIVENUMBER64
        | NUMBER64
        | HEX_STRING
        | BIN_STRING"""
        p[0] = p[1]

    def p_enumSpec(self, p):
        """enumSpec : '{' enumItems '}'"""
        p[0] = ("enumSpec", p[2])

    def p_enumItems(self, p):
        """enumItems : enumItems ',' enumItem
        | enumItem"""
        n = len(p)
        if n == 4:
            p[0] = p[1] + [p[3]]
        elif n == 2:
            p[0] = [p[1]]

    def p_enumItem(self, p):
        """enumItem : LOWERCASE_IDENTIFIER '(' enumNumber ')'"""
        p[0] = (p[1], p[3])

    def p_enumNumber(self, p):
        """enumNumber : NUMBER
        | NEGATIVENUMBER"""
        # XXX              | LOWERCASE_IDENTIFIER"""
        p[0] = p[1]

    def p_Status(self, p):
        """Status : LOWERCASE_IDENTIFIER"""
        p[0] = ("Status", p[1])

    def p_DisplayPart(self, p):
        """DisplayPart : DISPLAY_HINT Text
        | empty"""
        if p[1]:
            p[0] = (p[1], p[2])

    def p_UnitsPart(self, p):
        """UnitsPart : UNITS Text
        | empty"""
        if p[1]:
            p[0] = (p[1], p[2])

    def p_Access(self, p):
        """Access : LOWERCASE_IDENTIFIER"""
        p[0] = p[1]

    def p_IndexPart(self, p):
        """IndexPart : AUGMENTS '{' Entry '}'
        | empty"""
        if p[1]:
            p[0] = p[3]

    def p_MibIndex(self, p):
        """MibIndex : INDEX '{' IndexTypes '}'
        | empty"""
        if p[1]:
            p[0] = (p[1], p[3])

    def p_IndexTypes(self, p):
        """IndexTypes : IndexTypes ',' IndexType
        | IndexType"""
        n = len(p)
        if n == 4:
            p[0] = p[1] + [p[3]]
        elif n == 2:
            p[0] = [p[1]]

    def p_IndexType(self, p):
        """IndexType : IMPLIED Index
        | Index"""
        n = len(p)
        if n == 2:
            p[0] = (0, p[1])
        elif n == 3:
            p[0] = (1, p[2])  # IMPLIED

    def p_Index(self, p):
        """Index : ObjectName"""
        # libsmi: TODO: use the SYNTAX value of the correspondent
        #               OBJECT-TYPE invocation
        p[0] = p[1][1][0]  # XXX just name???

    def p_Entry(self, p):
        """Entry : ObjectName"""
        p[0] = p[1][1][0]

    def p_DefValPart(self, p):
        """DefValPart : DEFVAL '{' Value '}'
        | empty"""
        if p[1] and p[3] is not None:
            p[0] = (p[1], p[3])

    def p_Value(self, p):
        """Value : valueofObjectSyntax
        | '{' BitsValue '}'"""
        n = len(p)
        if n == 2:
            p[0] = p[1]
        elif n == 4:
            p[0] = p[2]

    def p_BitsValue(self, p):
        """BitsValue : BitNames
        | empty"""
        p[0] = p[1] or []

    def p_BitNames(self, p):
        """BitNames : BitNames ',' LOWERCASE_IDENTIFIER
        | LOWERCASE_IDENTIFIER"""
        n = len(p)
        if n == 4:
            p[0] = ("BitNames", p[1][1] + [p[3]])
        elif n == 2:
            p[0] = ("BitNames", [p[1]])

    def p_ObjectName(self, p):
        """ObjectName : objectIdentifier"""
        p[0] = p[1]

    def p_NotificationName(self, p):
        """NotificationName : objectIdentifier"""
        p[0] = p[1]

    def p_ReferPart(self, p):
        """ReferPart : REFERENCE Text
        | empty"""
        if p[1]:
            p[0] = (p[1], p[2])

    def p_RevisionPart(self, p):
        """RevisionPart : Revisions
        | empty"""
        if p[1]:
            p[0] = p[1]

    def p_Revisions(self, p):
        """Revisions : Revisions Revision
        | Revision"""
        n = len(p)
        if n == 3:
            p[0] = ("Revisions", p[1][1] + [p[2]])
        elif n == 2:
            p[0] = ("Revisions", [p[1]])

    def p_Revision(self, p):
        """Revision : REVISION ExtUTCTime DESCRIPTION Text"""
        p[0] = (p[2], (p[3], p[4]))  # revision time  # description

    def p_NotificationObjectsPart(self, p):
        """NotificationObjectsPart : OBJECTS '{' Objects '}'
        | empty"""
        p[0] = p[1] and p[3] or []

    def p_ObjectGroupObjectsPart(self, p):
        """ObjectGroupObjectsPart : OBJECTS '{' Objects '}'"""
        p[0] = p[3]

    def p_Objects(self, p):
        """Objects : Objects ',' Object
        | Object"""
        n = len(p)
        if n == 4:
            p[0] = ("Objects", p[1][1] + [p[3]])
        elif n == 2:
            p[0] = ("Objects", [p[1]])

    def p_Object(self, p):
        """Object : ObjectName"""
        p[0] = p[1][1][0]

    def p_NotificationsPart(self, p):
        """NotificationsPart : NOTIFICATIONS '{' Notifications '}'"""
        p[0] = p[3]

    def p_Notifications(self, p):
        """Notifications : Notifications ',' Notification
        | Notification"""
        n = len(p)
        if n == 4:
            p[0] = ("Notifications", p[1][1] + [p[3]])
        elif n == 2:
            p[0] = ("Notifications", [p[1]])

    def p_Notification(self, p):
        """Notification : NotificationName"""
        p[0] = p[1][1][0]

    def p_Text(self, p):
        """Text : QUOTED_STRING"""
        p[0] = p[1][1:-1]  # getting rid of quotes

    def p_ExtUTCTime(self, p):
        """ExtUTCTime : QUOTED_STRING"""
        p[0] = p[1][1:-1]  # getting rid of quotes

    def p_objectIdentifier(self, p):
        """objectIdentifier : subidentifiers"""
        p[0] = ("objectIdentifier", p[1])

    def p_subidentifiers(self, p):
        """subidentifiers : subidentifiers subidentifier
        | subidentifier"""
        n = len(p)
        if n == 3:
            p[0] = p[1] + [p[2]]
        elif n == 2:
            p[0] = [p[1]]

    def p_subidentifier(self, p):
        """subidentifier : fuzzy_lowercase_identifier
        | NUMBER
        | LOWERCASE_IDENTIFIER '(' NUMBER ')'"""
        n = len(p)
        if n == 2:
            p[0] = p[1]
        elif n == 5:
            # NOTE: we are not creating new symbol p[1] because formally
            # it is not defined in *this* MIB
            p[0] = (p[1], p[3])

    def p_objectIdentifier_defval(self, p):
        """objectIdentifier_defval : subidentifiers_defval"""
        p[0] = ("objectIdentifier_defval", p[1])

    def p_subidentifiers_defval(self, p):
        """subidentifiers_defval : subidentifiers_defval subidentifier_defval
        | subidentifier_defval"""
        n = len(p)
        if n == 3:
            p[0] = ("subidentifiers_defval", p[1][1] + [p[2]])
        elif n == 2:
            p[0] = ("subidentifiers_defval", [p[1]])

    def p_subidentifier_defval(self, p):
        """subidentifier_defval : LOWERCASE_IDENTIFIER '(' NUMBER ')'
        | NUMBER"""
        n = len(p)
        if n == 2:
            p[0] = ("subidentifier_defval", p[1])
        elif n == 5:
            p[0] = ("subidentifier_defval", p[1], p[3])

    def p_objectGroupClause(self, p):
        """objectGroupClause : LOWERCASE_IDENTIFIER OBJECT_GROUP ObjectGroupObjectsPart STATUS Status DESCRIPTION Text ReferPart COLON_COLON_EQUAL '{' objectIdentifier '}'"""
        p[0] = (
            "objectGroupClause",
            p[1],  # id
            p[3],  # objects
            p[5],  # status
            (p[6], p[7]),  # description
            p[8],  # reference
            p[11],
        )  # objectIdentifier

    def p_notificationGroupClause(self, p):
        """notificationGroupClause : LOWERCASE_IDENTIFIER NOTIFICATION_GROUP NotificationsPart STATUS Status DESCRIPTION Text ReferPart COLON_COLON_EQUAL '{' objectIdentifier '}'"""
        p[0] = (
            "notificationGroupClause",
            p[1],  # id
            p[3],  # notifications
            p[5],  # status
            (p[6], p[7]),  # description
            p[8],  # reference
            p[11],
        )  # objectIdentifier

    def p_moduleComplianceClause(self, p):
        """moduleComplianceClause : LOWERCASE_IDENTIFIER MODULE_COMPLIANCE STATUS Status DESCRIPTION Text ReferPart ComplianceModulePart COLON_COLON_EQUAL '{' objectIdentifier '}'"""
        p[0] = (
            "moduleComplianceClause",
            p[1],  # id
            #  p[2], # MODULE_COMPLIANCE
            p[4],  # status
            (p[5], p[6]),  # description
            p[7],  # reference
            p[8],  # ComplianceModules
            p[11],
        )  # objectIdentifier

    def p_ComplianceModulePart(self, p):
        """ComplianceModulePart : ComplianceModules"""
        p[0] = p[1]

    def p_ComplianceModules(self, p):
        """ComplianceModules : ComplianceModules ComplianceModule
        | ComplianceModule"""
        n = len(p)
        if n == 3:
            p[0] = ("ComplianceModules", p[1][1] + [p[2]])
        elif n == 2:
            p[0] = ("ComplianceModules", [p[1]])

    def p_ComplianceModule(self, p):
        """ComplianceModule : MODULE ComplianceModuleName MandatoryPart CompliancePart"""
        objects = p[3] and p[3][1] or []
        objects += p[4] and p[4][1] or []
        p[0] = (p[2], objects)  # ModuleName  # MandatoryPart + CompliancePart

    def p_ComplianceModuleName(self, p):
        """ComplianceModuleName : UPPERCASE_IDENTIFIER
        | empty"""
        # XXX                   | UPPERCASE_IDENTIFIER objectIdentifier
        p[0] = p[1]

    def p_MandatoryPart(self, p):
        """MandatoryPart : MANDATORY_GROUPS '{' MandatoryGroups '}'
        | empty"""
        if p[1]:
            p[0] = p[3]

    def p_MandatoryGroups(self, p):
        """MandatoryGroups : MandatoryGroups ',' MandatoryGroup
        | MandatoryGroup"""
        n = len(p)
        if n == 4:
            p[0] = ("MandatoryGroups", p[1][1] + [p[3]])
        elif n == 2:
            p[0] = ("MandatoryGroups", [p[1]])

    def p_MandatoryGroup(self, p):
        """MandatoryGroup : objectIdentifier"""
        p[0] = p[1][1][0]  # objectIdentifier? Maybe name?

    def p_CompliancePart(self, p):
        """CompliancePart : Compliances
        | empty"""
        if p[1]:
            p[0] = p[1]

    def p_Compliances(self, p):
        """Compliances : Compliances Compliance
        | Compliance"""
        n = len(p)
        if n == 3:
            p[0] = p[1] and p[2] and ("Compliances", p[1][1] + [p[2]]) or p[1]
        elif n == 2:
            p[0] = p[1] and ("Compliances", [p[1]]) or None

    def p_Compliance(self, p):
        """Compliance : ComplianceGroup
        | ComplianceObject"""
        if p[1]:
            p[0] = p[1]

    def p_ComplianceGroup(self, p):
        """ComplianceGroup : GROUP objectIdentifier DESCRIPTION Text"""
        p[0] = p[2][1][0]  # objectIdentifier
        #        p[1], # GROUP
        #        (p[3], p[4])) # description

    def p_ComplianceObject(self, p):
        """ComplianceObject : OBJECT ObjectName SyntaxPart WriteSyntaxPart AccessPart DESCRIPTION Text"""
        # p[0] = (p[1], # object
        #        p[2], # name
        #        p[3], # syntax
        #        p[4], # write syntax
        #        p[5], # access
        #        (p[6], p[7])) # description

    def p_SyntaxPart(self, p):
        """SyntaxPart : SYNTAX Syntax
        | empty"""
        if p[1]:
            p[0] = p[2]

    def p_WriteSyntaxPart(self, p):
        """WriteSyntaxPart : WRITE_SYNTAX WriteSyntax
        | empty"""
        if p[1]:
            p[0] = p[2]

    def p_WriteSyntax(self, p):
        """WriteSyntax : Syntax"""
        p[0] = ("WriteSyntax", p[1])

    def p_AccessPart(self, p):
        """AccessPart : MIN_ACCESS Access
        | empty"""
        if p[1]:
            p[0] = (p[1], p[2])

    def p_agentCapabilitiesClause(self, p):
        """agentCapabilitiesClause : LOWERCASE_IDENTIFIER AGENT_CAPABILITIES PRODUCT_RELEASE Text STATUS Status DESCRIPTION Text ReferPart ModulePart_Capabilities COLON_COLON_EQUAL '{' objectIdentifier '}'"""
        p[0] = (
            "agentCapabilitiesClause",
            p[1],  # id
            #   p[2], # AGENT_CAPABILITIES
            (p[3], p[4]),  # product release
            p[6],  # status
            (p[7], p[8]),  # description
            p[9],  # reference
            #   p[10], # module capabilities
            p[13],
        )  # objectIdentifier

    def p_ModulePart_Capabilities(self, p):
        """ModulePart_Capabilities : Modules_Capabilities
        | empty"""
        # if p[1]:
        #  p[0] = p[1]

    def p_Modules_Capabilities(self, p):
        """Modules_Capabilities : Modules_Capabilities Module_Capabilities
        | Module_Capabilities"""
        # n = len(p)
        # if n == 3:
        #  p[0] = ('Modules_Capabilities', p[1][1] + [p[2]])
        # elif n == 2:
        #  p[0] = ('Modules_Capabilities', [p[1]])

    def p_Module_Capabilities(self, p):
        """Module_Capabilities : SUPPORTS ModuleName_Capabilities INCLUDES '{' CapabilitiesGroups '}' VariationPart"""
        # p[0] = ('Module_Capabilities', (p[1], p[2]), # supports
        #                               (p[3], p[5]), # includes
        #                               p[7]) # variations

    def p_CapabilitiesGroups(self, p):
        """CapabilitiesGroups : CapabilitiesGroups ',' CapabilitiesGroup
        | CapabilitiesGroup"""
        # n = len(p)
        # if n == 4:
        #  p[0] = ('CapabilitiesGroups', p[1][1] + [p[3]])
        # elif n == 2:
        #  p[0] = ('CapabilitiesGroups', [p[1]])

    def p_CapabilitiesGroup(self, p):
        """CapabilitiesGroup : objectIdentifier"""
        # p[0] = ('CapabilitiesGroup', p[1])

    def p_ModuleName_Capabilities(self, p):
        """ModuleName_Capabilities : UPPERCASE_IDENTIFIER objectIdentifier
        | UPPERCASE_IDENTIFIER"""
        # n = len(p)
        # if n == 2:
        #  p[0] = ('ModuleName_Capabilities', p[1])
        # elif n == 3:
        #  p[0] = ('ModuleName_Capabilities', p[1], p[2])

    def p_VariationPart(self, p):
        """VariationPart : Variations
        | empty"""
        # if p[1]:
        #  p[0] = p[1]

    def p_Variations(self, p):
        """Variations : Variations Variation
        | Variation"""
        # n = len(p)
        # if n == 3:
        #  p[0] = ('Variations', p[1][1] + [p[2]])
        # elif n == 2:
        #  p[0] = ('Variations', [p[1]])        pass

    def p_Variation(self, p):
        """Variation : VARIATION ObjectName SyntaxPart WriteSyntaxPart VariationAccessPart CreationPart DefValPart DESCRIPTION Text"""
        # p[0] = (p[1], # variation
        #        p[2], # name
        #        p[3], # syntax
        #        p[4], # write syntax
        #        p[5], # access
        #        p[6], # creation
        #        p[7], # defval
        #        (p[8], p[9])) # description

    def p_VariationAccessPart(self, p):
        """VariationAccessPart : ACCESS VariationAccess
        | empty"""
        # if p[1]:
        #  p[0] = (p[1], p[2])

    def p_VariationAccess(self, p):
        """VariationAccess : LOWERCASE_IDENTIFIER"""
        # p[0] = p[1]

    def p_CreationPart(self, p):
        """CreationPart : CREATION_REQUIRES '{' Cells '}'
        | empty"""
        if p[1]:
            p[0] = (p[1], p[3])

    def p_Cells(self, p):
        """Cells : Cells ',' Cell
        | Cell"""
        n = len(p)
        if n == 4:
            p[0] = ("Cells", p[1][1] + [p[3]])
        elif n == 2:
            p[0] = ("Cells", [p[1]])

    def p_Cell(self, p):
        """Cell : ObjectName"""
        p[0] = ("Cell", p[1])

    def p_empty(self, p):
        """empty :"""

    # Error rule for syntax errors
    def p_error(self, p):
        if p:
            raise error.PySmiParserError(
                f"Bad grammar near token type {p.type}, value {p.value}",
                lineno=p.lineno,
            )


#
# Parser grammar relaxation follows.
#
# The classes that follow serve a purpose of encapsulating assorted functions
# into a namespace. The namespace type is not universally supported across all
# Python versions we want to run on, thus the hack with `staticmethod` decorator
# and `self` first parameter.
#

#
# SMIv1 grammar
#


# noinspection PyIncorrectDocstring
class SupportSmiV1Keywords:
    # NETWORKADDRESS added
    @staticmethod
    def p_importedKeyword(self, p):
        """importedKeyword : importedSMIKeyword
        | BITS
        | INTEGER32
        | IPADDRESS
        | NETWORKADDRESS
        | MANDATORY_GROUPS
        | MODULE_COMPLIANCE
        | MODULE_IDENTITY
        | OBJECT_GROUP
        | OBJECT_IDENTITY
        | OBJECT_TYPE
        | OPAQUE
        | TEXTUAL_CONVENTION
        | TIMETICKS
        | UNSIGNED32"""
        p[0] = p[1]

    # NETWORKADDRESS added
    @staticmethod
    def p_typeSMIandSPPI(self, p):
        """typeSMIandSPPI : IPADDRESS
        | NETWORKADDRESS
        | TIMETICKS
        | OPAQUE
        | INTEGER32
        | UNSIGNED32"""
        p[0] = p[1]

    # NETWORKADDRESS added
    @staticmethod
    def p_ApplicationSyntax(self, p):
        """ApplicationSyntax : IPADDRESS anySubType
        | NETWORKADDRESS anySubType
        | COUNTER32
        | COUNTER32 integerSubType
        | GAUGE32
        | GAUGE32 integerSubType
        | UNSIGNED32
        | UNSIGNED32 integerSubType
        | TIMETICKS anySubType
        | OPAQUE
        | OPAQUE octetStringSubType
        | COUNTER64
        | COUNTER64 integerSubType"""
        n = len(p)
        if n == 2:
            p[0] = ("ApplicationSyntax", p[1])
        elif n == 3:
            p[0] = ("ApplicationSyntax", p[1], p[2])

    # NETWORKADDRESS added for SEQUENCE syntax
    @staticmethod
    def p_sequenceApplicationSyntax(self, p):
        """sequenceApplicationSyntax : IPADDRESS anySubType
        | NETWORKADDRESS anySubType
        | COUNTER32 anySubType
        | GAUGE32 anySubType
        | UNSIGNED32 anySubType
        | TIMETICKS anySubType
        | OPAQUE
        | COUNTER64 anySubType"""
        n = len(p)
        if n == 2:
            p[0] = p[1]
        elif n == 3:
            p[0] = p[1]  # XXX not supporting subtypes here


# noinspection PyIncorrectDocstring
class SupportIndex:
    # SMIv1 IndexTypes added
    @staticmethod
    def p_Index(self, p):
        """Index : ObjectName
        | typeSMIv1"""

        # libsmi: TODO: use the SYNTAX value of the correspondent
        #               OBJECT-TYPE invocation
        p[0] = isinstance(p[1], tuple) and p[1][1][0] or p[1]

    # for Index rule
    @staticmethod
    def p_typeSMIv1(self, p):
        """typeSMIv1 : INTEGER
        | OCTET STRING
        | IPADDRESS
        | NETWORKADDRESS"""
        n = len(p)
        indextype = n == 3 and p[1] + " " + p[2] or p[1]
        p[0] = indextype


#
# Some changes in grammar to handle common mistakes in MIBs
#


# noinspection PyIncorrectDocstring
class CommaInImport:
    # comma at the end of import list
    @staticmethod
    def p_importIdentifiers(self, p):
        """importIdentifiers : importIdentifiers ',' importIdentifier
        | importIdentifier
        | importIdentifiers ','"""
        n = len(p)
        if n == 4:
            p[0] = p[1] + [p[3]]
        elif n == 2:
            p[0] = [p[1]]
        elif n == 3:  # excessive comma case
            p[0] = p[1]


# noinspection PyIncorrectDocstring
class CommaInSequence:
    # comma at the end of sequence list
    @staticmethod
    def p_sequenceItems(self, p):
        """sequenceItems : sequenceItems ',' sequenceItem
        | sequenceItem
        | sequenceItems ','"""
        # libsmi: TODO: might this list be emtpy?
        n = len(p)
        if n == 4:
            p[0] = p[1] + [p[3]]
        elif n == 2:
            p[0] = [p[1]]
        elif n == 3:  # excessive comma case
            p[0] = p[1]


# noinspection PyIncorrectDocstring
class CommaAndSpaces:
    # common typos handled (mix of commas and spaces)
    @staticmethod
    def p_enumItems(self, p):
        """enumItems : enumItems ',' enumItem
        | enumItem
        | enumItems enumItem
        | enumItems ','"""
        n = len(p)
        if n == 4:
            p[0] = p[1] + [p[3]]
        elif n == 2:
            p[0] = [p[1]]
        elif n == 3:  # typo case
            if p[2] == ",":
                p[0] = p[1]
            else:
                p[0] = p[1] + [p[2]]


# noinspection PyIncorrectDocstring
class UppercaseIdentifier:
    # common mistake - using UPPERCASE_IDENTIFIER
    @staticmethod
    def p_enumItem(self, p):
        """enumItem : LOWERCASE_IDENTIFIER '(' enumNumber ')'
        | UPPERCASE_IDENTIFIER '(' enumNumber ')'"""
        p[0] = (p[1], p[3])


# noinspection PyIncorrectDocstring
class LowcaseIdentifier:
    # common mistake - LOWERCASE_IDENTIFIER in symbol's name
    @staticmethod
    def p_notificationTypeClause(self, p):
        """notificationTypeClause : fuzzy_lowercase_identifier NOTIFICATION_TYPE NotificationObjectsPart STATUS Status DESCRIPTION Text ReferPart COLON_COLON_EQUAL '{' NotificationName '}'"""  # some MIBs have uppercase and/or lowercase id
        p[0] = (
            "notificationTypeClause",
            p[1],  # id
            #  p[2], # NOTIFICATION_TYPE
            p[3],  # NotificationObjectsPart
            p[5],  # status
            (p[6], p[7]),  # description
            p[8],  # Reference
            p[11],
        )  # NotificationName aka objectIdentifier


# noinspection PyIncorrectDocstring,PyIncorrectDocstring
class CurlyBracesInEnterprises:
    # common mistake - curly brackets around enterprise symbol
    @staticmethod
    def p_trapTypeClause(self, p):
        """trapTypeClause : fuzzy_lowercase_identifier TRAP_TYPE EnterprisePart VarPart DescrPart ReferPart COLON_COLON_EQUAL NUMBER"""
        # libsmi: TODO: range of number?
        p[0] = (
            "trapTypeClause",
            p[1],  # fuzzy_lowercase_identifier
            #  p[2], # TRAP_TYPE
            p[3],  # EnterprisePart (objectIdentifier)
            p[4],  # VarPart
            p[5],  # description
            p[6],  # reference
            p[8],
        )  # NUMBER

    @staticmethod
    def p_EnterprisePart(self, p):
        """EnterprisePart : ENTERPRISE objectIdentifier
        | ENTERPRISE '{' objectIdentifier '}'"""
        n = len(p)
        if n == 3:
            p[0] = p[2]
        elif n == 5:  # common mistake case
            p[0] = p[3]


# noinspection PyIncorrectDocstring
class NoCells:
    # common mistake - no Cells
    @staticmethod
    def p_CreationPart(self, p):
        """CreationPart : CREATION_REQUIRES '{' Cells '}'
        | CREATION_REQUIRES '{' '}'
        | empty"""
        n = len(p)
        if n == 5:
            p[0] = (p[1], p[3])


relaxedGrammar = {
    "supportSmiV1Keywords": [
        SupportSmiV1Keywords.p_importedKeyword,
        SupportSmiV1Keywords.p_typeSMIandSPPI,
        SupportSmiV1Keywords.p_ApplicationSyntax,
        SupportSmiV1Keywords.p_sequenceApplicationSyntax,
    ],
    "supportIndex": [SupportIndex.p_Index, SupportIndex.p_typeSMIv1],
    "commaAtTheEndOfImport": [CommaInImport.p_importIdentifiers],
    "commaAtTheEndOfSequence": [CommaInSequence.p_sequenceItems],
    "mixOfCommasAndSpaces": [CommaAndSpaces.p_enumItems],
    "uppercaseIdentifier": [UppercaseIdentifier.p_enumItem],
    "lowcaseIdentifier": [LowcaseIdentifier.p_notificationTypeClause],
    "curlyBracesAroundEnterpriseInTrap": [
        CurlyBracesInEnterprises.p_trapTypeClause,
        CurlyBracesInEnterprises.p_EnterprisePart,
    ],
    "noCells": [NoCells.p_CreationPart],
}


def parserFactory(**grammarOptions):
    """Factory function producing custom specializations of base *SmiV2Parser*
    class.

    Keyword Args:
        grammarOptions: a list of (bool) typed optional keyword parameters
                        enabling particular set of SMIv2 grammar relaxations.

    Returns:
        Specialized copy of *SmiV2Parser* class.

    Notes:
        The following SMIv2 grammar relaxation parameters are defined:

        * supportSmiV1Keywords - parses SMIv1 grammar
        * supportIndex - tolerates ASN.1 types in INDEX clause
        * commaAtTheEndOfImport - tolerates stray comma at the end of IMPORT section
        * commaAtTheEndOfSequence - tolerates stray comma at the end of sequence of elements in MIB
        * mixOfCommasAndSpaces - tolerate a mix of comma and spaces in MIB enumerations
        * uppercaseIdentifier - tolerate uppercased MIB identifiers
        * lowcaseIdentifier - tolerate lowercase MIB identifiers
        * curlyBracesAroundEnterpriseInTrap - tolerate curly braces around enterprise ID in TRAP MACRO
        * noCells - tolerate missing cells (XXX)

    Examples:

    >>> from pysmi.parser import smi
    >>> SmiV1Parser = smi.parserFactory(supportSmiV1Keywords=True, supportIndex=True)

    """
    classAttr = {}

    for option in grammarOptions:
        if grammarOptions[option]:
            if option not in relaxedGrammar:
                raise error.PySmiError(f"Unknown parser relaxation option: {option}")

            for func in relaxedGrammar[option]:
                classAttr[func.__name__] = func

    classAttr["defaultLexer"] = lexerFactory(**grammarOptions)

    return type("SmiParser", (SmiV2Parser,), classAttr)



#!/usr/bin/env python
#
# This file is part of pysmi software.
#
# Copyright (c) 2015-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysmi/license.html
#
# SNMP SMI/MIB copying tool
#
import getopt
import os
import shutil
import sys
from datetime import datetime

from pysmi import debug, error
from pysmi.codegen import JsonCodeGen
from pysmi.compiler import MibCompiler
from pysmi.parser import SmiV1CompatParser
from pysmi.reader import FileReader, get_readers_from_urls
from pysmi.writer import CallbackWriter


def start():
    # sysexits.h
    EX_OK = 0
    EX_USAGE = 64
    EX_SOFTWARE = 70

    # Defaults
    quietFlag = False
    verboseFlag = False
    mibSources = []
    dstDirectory = None
    cacheDirectory = ""
    dryrunFlag = False
    ignoreErrorsFlag = False

    helpMessage = f"""\
    Usage: {sys.argv[0]} [--help]
        [--version]
        [--verbose]
        [--quiet]
        [--debug=<{"|".join(sorted(debug.FLAG_MAP))}>]
        [--mib-source=<URI>]
        [--cache-directory=<DIRECTORY>]
        [--ignore-errors]
        [--dry-run]
        <SOURCE [SOURCE...]> <DESTINATION>
    Where:
        URI      - file, zip, http, https schemes are supported.
                Use @mib@ placeholder token in URI to refer directly to
                the required MIB module when source does not support
                directory listing (e.g. HTTP).
    """

    # TODO(etingof): add the option to copy MIBs into enterprise-indexed subdirs

    try:
        opts, inputMibs = getopt.getopt(
            sys.argv[1:],
            "hv",
            [
                "help",
                "version",
                "verbose",
                "quiet",
                "debug=",
                "mib-source=",
                "mib-stub=",
                "cache-directory=",
                "ignore-errors",
                "dry-run",
            ],
        )

    except getopt.GetoptError:
        sys.exit(EX_USAGE)

    for opt in opts:
        if opt[0] == "-h" or opt[0] == "--help":
            sys.stderr.write(
                f"""\
    Synopsis:
    SNMP SMI/MIB files copying tool. When given MIB file(s) or directory(ies)
    on input and a destination directory, the tool parses MIBs to figure out
    their canonical MIB module name and the latest revision date, then
    copies MIB module on input into the destination directory under its
    MIB module name *if* there is no such file already or its revision date
    is older.

    Documentation:
    https://www.pysnmp.com/pysmi
    {helpMessage}
    """
            )
            sys.exit(EX_OK)

        if opt[0] == "-v" or opt[0] == "--version":
            from pysmi import __version__

            sys.stderr.write(
                f"""\
    SNMP SMI/MIB library version {__version__}, written by Ilya Etingof <etingof@gmail.com>
    Python interpreter: {sys.version}
    Software documentation and support at https://www.pysnmp.com/pysmi
    {helpMessage}
    """
            )
            sys.exit(EX_OK)

        if opt[0] == "--quiet":
            quietFlag = True

        if opt[0] == "--verbose":
            verboseFlag = True

        if opt[0] == "--debug":
            debug.set_logger(debug.Debug(*opt[1].split(",")))

        if opt[0] == "--mib-source":
            mibSources.append(opt[1])

        if opt[0] == "--cache-directory":
            cacheDirectory = opt[1]

        if opt[0] == "--ignore-errors":
            ignoreErrorsFlag = True

        if opt[0] == "--dry-run":
            dryrunFlag = True

    if not mibSources:
        mibSources = [
            "file:///usr/share/snmp/mibs",
            "https://mibs.pysnmp.com/asn1/@mib@",
        ]

    if len(inputMibs) < 2:
        sys.stderr.write(
            f"ERROR: MIB source and/or destination arguments not given{os.linesep}{helpMessage}{os.linesep}"
        )
        sys.exit(EX_USAGE)

    dstDirectory = inputMibs.pop()

    if os.path.exists(dstDirectory) and not os.path.isdir(dstDirectory):
        sys.stderr.write(
            f"ERROR: given destination '{dstDirectory}' is not a directory{os.linesep}{helpMessage}{os.linesep}"
        )
        sys.exit(EX_USAGE)

    try:
        os.makedirs(dstDirectory, mode=0o755)

    except OSError:
        pass

    # Compiler infrastructure

    codeGenerator = JsonCodeGen()

    mibParser = SmiV1CompatParser(tempdir=cacheDirectory)

    fileWriter = CallbackWriter(lambda *x: None)

    def get_mib_revision(mibDir, mibFile):
        mibCompiler = MibCompiler(mibParser, codeGenerator, fileWriter)

        mibCompiler.add_sources(
            FileReader(mibDir, recursive=False, ignoreErrors=ignoreErrorsFlag),
            *get_readers_from_urls(*mibSources),
        )

        try:
            processed = mibCompiler.compile(
                mibFile,
                **dict(
                    noDeps=True,
                    rebuild=True,
                    fuzzyMatching=False,
                    ignoreErrors=ignoreErrorsFlag,
                ),
            )

        except error.PySmiError:
            sys.stderr.write(f"ERROR: {sys.exc_info()[1]}{os.linesep}")
            sys.exit(EX_SOFTWARE)

        for canonicalMibName in processed:
            if processed[canonicalMibName] == "compiled" and processed[
                canonicalMibName
            ].path == "file://" + os.path.join(mibDir, mibFile):
                try:
                    revision = datetime.strptime(
                        processed[canonicalMibName].revision, "%Y-%m-%d %H:%M"
                    )

                except Exception:
                    revision = datetime.fromtimestamp(0)

                return canonicalMibName, revision

        raise error.PySmiError(
            f'Can\'t read or parse MIB "{os.path.join(mibDir, mibFile)}"'
        )

    def shorten_path(path, maxLength=45):
        if len(path) > maxLength:
            return "..." + path[-maxLength:]
        else:
            return path

    mibsSeen = mibsCopied = mibsFailed = 0

    mibsRevisions = {}

    for srcDirectory in inputMibs:
        if verboseFlag:
            if verboseFlag:
                sys.stderr.write(f'Reading "{srcDirectory}"...{os.linesep}')

        if os.path.isfile(srcDirectory):
            mibFiles = [
                (
                    os.path.abspath(os.path.dirname(srcDirectory)),
                    os.path.basename(srcDirectory),
                )
            ]

        else:
            mibFiles = [
                (os.path.abspath(dirName), mibFile)
                for dirName, _, mibFiles in os.walk(srcDirectory)
                for mibFile in mibFiles
            ]

        for srcDirectory, mibFile in mibFiles:
            mibsSeen += 1

            # TODO(etingof): also check module OID to make sure there is no name collision

            try:
                mibName, srcMibRevision = get_mib_revision(srcDirectory, mibFile)

            except error.PySmiError as ex:
                if verboseFlag:
                    sys.stderr.write(
                        f'Failed to read source MIB "{os.path.join(srcDirectory, mibFile)}": {ex}{os.linesep}'
                    )

                if not quietFlag:
                    sys.stderr.write(
                        f"FAILED {shorten_path(os.path.join(srcDirectory, mibFile))}{os.linesep}"
                    )

                mibsFailed += 1

                continue

            if mibName in mibsRevisions:
                dstMibRevision = mibsRevisions[mibName]

            else:
                try:
                    _, dstMibRevision = get_mib_revision(dstDirectory, mibName)

                except error.PySmiError as ex:
                    if verboseFlag:
                        sys.stderr.write(
                            f'MIB "{os.path.join(srcDirectory, mibFile)}" is not available at the destination directory "{dstDirectory}": {ex}{os.linesep}'
                        )

                    dstMibRevision = datetime.fromtimestamp(0)

                mibsRevisions[mibName] = dstMibRevision

            if dstMibRevision >= srcMibRevision:
                if verboseFlag:
                    sys.stderr.write(
                        f'Destination MIB "{os.path.join(dstDirectory, mibName)}" has the same or newer revision as the source MIB "{os.path.join(srcDirectory, mibFile)}"{os.linesep}'
                    )
                if not quietFlag:
                    sys.stderr.write(
                        f"NOT COPIED {shorten_path(os.path.join(srcDirectory, mibFile))} ({mibName}){os.linesep}"
                    )

                continue

            mibsRevisions[mibName] = srcMibRevision

            if verboseFlag:
                if verboseFlag:
                    sys.stderr.write(
                        f'Copying "{os.path.join(srcDirectory, mibFile)}" (revision "{srcMibRevision}") -> "{os.path.join(dstDirectory, mibName)}" (revision "{dstMibRevision}"){os.linesep}'
                    )

            try:
                shutil.copy(
                    os.path.join(srcDirectory, mibFile),
                    os.path.join(dstDirectory, mibName),
                )

            except Exception as ex:
                if verboseFlag:
                    sys.stderr.write(
                        f'Failed to copy MIB "{os.path.join(srcDirectory, mibFile)}" -> "{os.path.join(dstDirectory, mibName)}" ({mibName}): "{ex}"{os.linesep}'
                    )

                if not quietFlag:
                    sys.stderr.write(
                        f"FAILED {shorten_path(os.path.join(srcDirectory, mibFile))} ({mibName}){os.linesep}"
                    )

                mibsFailed += 1

            else:
                if not quietFlag:
                    sys.stderr.write(
                        f"COPIED {shorten_path(os.path.join(srcDirectory, mibFile))} ({mibName}){os.linesep}"
                    )

                mibsCopied += 1

    if not quietFlag:
        sys.stderr.write(
            f"MIBs seen: {mibsSeen}, copied: {mibsCopied}, failed: {mibsFailed}{os.linesep}"
        )

    sys.exit(EX_OK)



#
# This file is part of pysmi software.
#
# Copyright (c) 2015-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysmi/license.html
#
from keyword import iskeyword
import warnings

from pysmi import error

# Prefix added to symbols that happen to be Python keywords. The leading
# underscore ensures that there is no overlap with any other symbols.
RESERVED_KEYWORDS_PREFIX = "_pysmi_"


def dorepr(s):
    return repr(s)


def update_dict(d1, d2):
    d1.update(d2)
    return d1


class AbstractCodeGen:
    # never compile these, they either:
    # - define MACROs (implementation supplies them)
    # - or carry conflicting OIDs (so that all IMPORT's of them will be rewritten)
    # - or have manual fixes
    # - or import base ASN.1 types from implementation-specific MIBs
    baseMibs = (
        "RFC1065-SMI",
        "RFC1155-SMI",
        "RFC1158-MIB",
        "RFC-1212",
        "RFC1213-MIB",
        "RFC-1215",
        "SNMPv2-SMI",
        "SNMPv2-TC",
        "SNMPv2-TM",
        "SNMPv2-CONF",
    )

    # Explicit SMIv1 -> SMIv2 mapping for standard MIBs
    commonSyms = {
        "RFC1155-SMI/RFC1065-SMI": {
            "internet": [("SNMPv2-SMI", "internet")],
            "directory": [("SNMPv2-SMI", "directory")],
            "mgmt": [("SNMPv2-SMI", "mgmt")],
            "experimental": [("SNMPv2-SMI", "experimental")],
            "private": [("SNMPv2-SMI", "private")],
            "enterprises": [("SNMPv2-SMI", "enterprises")],
            "OBJECT-TYPE": [("SNMPv2-SMI", "OBJECT-TYPE")],
            "ObjectName": [("SNMPv2-SMI", "ObjectName")],
            "ObjectSyntax": [("SNMPv2-SMI", "ObjectSyntax")],
            "SimpleSyntax": [("SNMPv2-SMI", "SimpleSyntax")],
            "ApplicationSyntax": [("SNMPv2-SMI", "ApplicationSyntax")],
            "NetworkAddress": [("SNMPv2-SMI", "IpAddress")],
            "IpAddress": [("SNMPv2-SMI", "IpAddress")],
            "Counter": [("SNMPv2-SMI", "Counter32")],
            "Gauge": [("SNMPv2-SMI", "Gauge32")],
            "TimeTicks": [("SNMPv2-SMI", "TimeTicks")],
            "Opaque": [("SNMPv2-SMI", "Opaque")],
        },
        "RFC1158-MIB/RFC1213-MIB": {
            "mib-2": [("SNMPv2-SMI", "mib-2")],
            "DisplayString": [("SNMPv2-TC", "DisplayString")],
            "system": [("SNMPv2-MIB", "system")],
            "interfaces": [("IF-MIB", "interfaces")],
            "ip": [("IP-MIB", "ip")],
            "icmp": [("IP-MIB", "icmp")],
            "tcp": [("TCP-MIB", "tcp")],
            "udp": [("UDP-MIB", "udp")],
            "transmission": [("SNMPv2-SMI", "transmission")],
            "snmp": [("SNMPv2-MIB", "snmp")],
            "sysDescr": [("SNMPv2-MIB", "sysDescr")],
            "sysObjectID": [("SNMPv2-MIB", "sysObjectID")],
            "sysUpTime": [("SNMPv2-MIB", "sysUpTime")],
            "sysContact": [("SNMPv2-MIB", "sysContact")],
            "sysName": [("SNMPv2-MIB", "sysName")],
            "sysLocation": [("SNMPv2-MIB", "sysLocation")],
            "sysServices": [("SNMPv2-MIB", "sysServices")],
            "ifNumber": [("IF-MIB", "ifNumber")],
            "ifTable": [("IF-MIB", "ifTable")],
            "ifEntry": [("IF-MIB", "ifEntry")],
            "ifIndex": [("IF-MIB", "ifIndex")],
            "ifDescr": [("IF-MIB", "ifDescr")],
            "ifType": [("IF-MIB", "ifType")],
            "ifMtu": [("IF-MIB", "ifMtu")],
            "ifSpeed": [("IF-MIB", "ifSpeed")],
            "ifPhysAddress": [("IF-MIB", "ifPhysAddress")],
            "ifAdminStatus": [("IF-MIB", "ifAdminStatus")],
            "ifOperStatus": [("IF-MIB", "ifOperStatus")],
            "ifLastChange": [("IF-MIB", "ifLastChange")],
            "ifInOctets": [("IF-MIB", "ifInOctets")],
            "ifInUcastPkts": [("IF-MIB", "ifInUcastPkts")],
            "ifInNUcastPkts": [("IF-MIB", "ifInNUcastPkts")],
            "ifInDiscards": [("IF-MIB", "ifInDiscards")],
            "ifInErrors": [("IF-MIB", "ifInErrors")],
            "ifInUnknownProtos": [("IF-MIB", "ifInUnknownProtos")],
            "ifOutOctets": [("IF-MIB", "ifOutOctets")],
            "ifOutUcastPkts": [("IF-MIB", "ifOutUcastPkts")],
            "ifOutNUcastPkts": [("IF-MIB", "ifOutNUcastPkts")],
            "ifOutDiscards": [("IF-MIB", "ifOutDiscards")],
            "ifOutErrors": [("IF-MIB", "ifOutErrors")],
            "ifOutQLen": [("IF-MIB", "ifOutQLen")],
            "ifSpecific": [("IF-MIB", "ifSpecific")],
            "ipForwarding": [("IP-MIB", "ipForwarding")],
            "ipDefaultTTL": [("IP-MIB", "ipDefaultTTL")],
            "ipInReceives": [("IP-MIB", "ipInReceives")],
            "ipInHdrErrors": [("IP-MIB", "ipInHdrErrors")],
            "ipInAddrErrors": [("IP-MIB", "ipInAddrErrors")],
            "ipForwDatagrams": [("IP-MIB", "ipForwDatagrams")],
            "ipInUnknownProtos": [("IP-MIB", "ipInUnknownProtos")],
            "ipInDiscards": [("IP-MIB", "ipInDiscards")],
            "ipInDelivers": [("IP-MIB", "ipInDelivers")],
            "ipOutRequests": [("IP-MIB", "ipOutRequests")],
            "ipOutDiscards": [("IP-MIB", "ipOutDiscards")],
            "ipOutNoRoutes": [("IP-MIB", "ipOutNoRoutes")],
            "ipReasmTimeout": [("IP-MIB", "ipReasmTimeout")],
            "ipReasmReqds": [("IP-MIB", "ipReasmReqds")],
            "ipReasmOKs": [("IP-MIB", "ipReasmOKs")],
            "ipReasmFails": [("IP-MIB", "ipReasmFails")],
            "ipFragOKs": [("IP-MIB", "ipFragOKs")],
            "ipFragFails": [("IP-MIB", "ipFragFails")],
            "ipFragCreates": [("IP-MIB", "ipFragCreates")],
            "ipAddrTable": [("IP-MIB", "ipAddrTable")],
            "ipAddrEntry": [("IP-MIB", "ipAddrEntry")],
            "ipAdEntAddr": [("IP-MIB", "ipAdEntAddr")],
            "ipAdEntIfIndex": [("IP-MIB", "ipAdEntIfIndex")],
            "ipAdEntNetMask": [("IP-MIB", "ipAdEntNetMask")],
            "ipAdEntBcastAddr": [("IP-MIB", "ipAdEntBcastAddr")],
            "ipAdEntReasmMaxSize": [("IP-MIB", "ipAdEntReasmMaxSize")],
            "ipNetToMediaTable": [("IP-MIB", "ipNetToMediaTable")],
            "ipNetToMediaEntry": [("IP-MIB", "ipNetToMediaEntry")],
            "ipNetToMediaIfIndex": [("IP-MIB", "ipNetToMediaIfIndex")],
            "ipNetToMediaPhysAddress": [("IP-MIB", "ipNetToMediaPhysAddress")],
            "ipNetToMediaNetAddress": [("IP-MIB", "ipNetToMediaNetAddress")],
            "ipNetToMediaType": [("IP-MIB", "ipNetToMediaType")],
            "icmpInMsgs": [("IP-MIB", "icmpInMsgs")],
            "icmpInErrors": [("IP-MIB", "icmpInErrors")],
            "icmpInDestUnreachs": [("IP-MIB", "icmpInDestUnreachs")],
            "icmpInTimeExcds": [("IP-MIB", "icmpInTimeExcds")],
            "icmpInParmProbs": [("IP-MIB", "icmpInParmProbs")],
            "icmpInSrcQuenchs": [("IP-MIB", "icmpInSrcQuenchs")],
            "icmpInRedirects": [("IP-MIB", "icmpInRedirects")],
            "icmpInEchos": [("IP-MIB", "icmpInEchos")],
            "icmpInEchoReps": [("IP-MIB", "icmpInEchoReps")],
            "icmpInTimestamps": [("IP-MIB", "icmpInTimestamps")],
            "icmpInTimestampReps": [("IP-MIB", "icmpInTimestampReps")],
            "icmpInAddrMasks": [("IP-MIB", "icmpInAddrMasks")],
            "icmpInAddrMaskReps": [("IP-MIB", "icmpInAddrMaskReps")],
            "icmpOutMsgs": [("IP-MIB", "icmpOutMsgs")],
            "icmpOutErrors": [("IP-MIB", "icmpOutErrors")],
            "icmpOutDestUnreachs": [("IP-MIB", "icmpOutDestUnreachs")],
            "icmpOutTimeExcds": [("IP-MIB", "icmpOutTimeExcds")],
            "icmpOutParmProbs": [("IP-MIB", "icmpOutParmProbs")],
            "icmpOutSrcQuenchs": [("IP-MIB", "icmpOutSrcQuenchs")],
            "icmpOutRedirects": [("IP-MIB", "icmpOutRedirects")],
            "icmpOutEchos": [("IP-MIB", "icmpOutEchos")],
            "icmpOutEchoReps": [("IP-MIB", "icmpOutEchoReps")],
            "icmpOutTimestamps": [("IP-MIB", "icmpOutTimestamps")],
            "icmpOutTimestampReps": [("IP-MIB", "icmpOutTimestampReps")],
            "icmpOutAddrMasks": [("IP-MIB", "icmpOutAddrMasks")],
            "icmpOutAddrMaskReps": [("IP-MIB", "icmpOutAddrMaskReps")],
            "tcpRtoAlgorithm": [("TCP-MIB", "tcpRtoAlgorithm")],
            "tcpRtoMin": [("TCP-MIB", "tcpRtoMin")],
            "tcpRtoMax": [("TCP-MIB", "tcpRtoMax")],
            "tcpMaxConn": [("TCP-MIB", "tcpMaxConn")],
            "tcpActiveOpens": [("TCP-MIB", "tcpActiveOpens")],
            "tcpPassiveOpens": [("TCP-MIB", "tcpPassiveOpens")],
            "tcpAttemptFails": [("TCP-MIB", "tcpAttemptFails")],
            "tcpEstabResets": [("TCP-MIB", "tcpEstabResets")],
            "tcpCurrEstab": [("TCP-MIB", "tcpCurrEstab")],
            "tcpInSegs": [("TCP-MIB", "tcpInSegs")],
            "tcpOutSegs": [("TCP-MIB", "tcpOutSegs")],
            "tcpRetransSegs": [("TCP-MIB", "tcpRetransSegs")],
            "tcpConnTable": [("TCP-MIB", "tcpConnTable")],
            "tcpConnEntry": [("TCP-MIB", "tcpConnEntry")],
            "tcpConnState": [("TCP-MIB", "tcpConnState")],
            "tcpConnLocalAddress": [("TCP-MIB", "tcpConnLocalAddress")],
            "tcpConnLocalPort": [("TCP-MIB", "tcpConnLocalPort")],
            "tcpConnRemAddress": [("TCP-MIB", "tcpConnRemAddress")],
            "tcpConnRemPort": [("TCP-MIB", "tcpConnRemPort")],
            "tcpInErrs": [("TCP-MIB", "tcpInErrs")],
            "tcpOutRsts": [("TCP-MIB", "tcpOutRsts")],
            "udpInDatagrams": [("UDP-MIB", "udpInDatagrams")],
            "udpNoPorts": [("UDP-MIB", "udpNoPorts")],
            "udpInErrors": [("UDP-MIB", "udpInErrors")],
            "udpOutDatagrams": [("UDP-MIB", "udpOutDatagrams")],
            "udpTable": [("UDP-MIB", "udpTable")],
            "udpEntry": [("UDP-MIB", "udpEntry")],
            "udpLocalAddress": [("UDP-MIB", "udpLocalAddress")],
            "udpLocalPort": [("UDP-MIB", "udpLocalPort")],
            "snmpInPkts": [("SNMPv2-MIB", "snmpInPkts")],
            "snmpOutPkts": [("SNMPv2-MIB", "snmpOutPkts")],
            "snmpInBadVersions": [("SNMPv2-MIB", "snmpInBadVersions")],
            "snmpInBadCommunityNames": [("SNMPv2-MIB", "snmpInBadCommunityNames")],
            "snmpInBadCommunityUses": [("SNMPv2-MIB", "snmpInBadCommunityUses")],
            "snmpInASNParseErrs": [("SNMPv2-MIB", "snmpInASNParseErrs")],
            "snmpInTooBigs": [("SNMPv2-MIB", "snmpInTooBigs")],
            "snmpInNoSuchNames": [("SNMPv2-MIB", "snmpInNoSuchNames")],
            "snmpInBadValues": [("SNMPv2-MIB", "snmpInBadValues")],
            "snmpInReadOnlys": [("SNMPv2-MIB", "snmpInReadOnlys")],
            "snmpInGenErrs": [("SNMPv2-MIB", "snmpInGenErrs")],
            "snmpInTotalReqVars": [("SNMPv2-MIB", "snmpInTotalReqVars")],
            "snmpInTotalSetVars": [("SNMPv2-MIB", "snmpInTotalSetVars")],
            "snmpInGetRequests": [("SNMPv2-MIB", "snmpInGetRequests")],
            "snmpInGetNexts": [("SNMPv2-MIB", "snmpInGetNexts")],
            "snmpInSetRequests": [("SNMPv2-MIB", "snmpInSetRequests")],
            "snmpInGetResponses": [("SNMPv2-MIB", "snmpInGetResponses")],
            "snmpInTraps": [("SNMPv2-MIB", "snmpInTraps")],
            "snmpOutTooBigs": [("SNMPv2-MIB", "snmpOutTooBigs")],
            "snmpOutNoSuchNames": [("SNMPv2-MIB", "snmpOutNoSuchNames")],
            "snmpOutBadValues": [("SNMPv2-MIB", "snmpOutBadValues")],
            "snmpOutGenErrs": [("SNMPv2-MIB", "snmpOutGenErrs")],
            "snmpOutGetRequests": [("SNMPv2-MIB", "snmpOutGetRequests")],
            "snmpOutGetNexts": [("SNMPv2-MIB", "snmpOutGetNexts")],
            "snmpOutSetRequests": [("SNMPv2-MIB", "snmpOutSetRequests")],
            "snmpOutGetResponses": [("SNMPv2-MIB", "snmpOutGetResponses")],
            "snmpOutTraps": [("SNMPv2-MIB", "snmpOutTraps")],
            "snmpEnableAuthenTraps": [("SNMPv2-MIB", "snmpEnableAuthenTraps")],
        },
    }

    convertImportv2 = {
        "RFC1065-SMI": commonSyms["RFC1155-SMI/RFC1065-SMI"],
        "RFC1155-SMI": commonSyms["RFC1155-SMI/RFC1065-SMI"],
        "RFC1158-MIB": update_dict(
            dict(commonSyms["RFC1158-MIB/RFC1213-MIB"]),
            (
                ("nullSpecific", [("SNMPv2-SMI", "zeroDotZero")]),
                ("ipRoutingTable", [("RFC1213-MIB", "ipRouteTable")]),
                ("ipRouteEntry", [("RFC1213-MIB", "ipRouteEntry")]),
                ("ipRouteDest", [("RFC1213-MIB", "ipRouteDest")]),
                ("ipRouteIfIndex", [("RFC1213-MIB", "ipRouteIfIndex")]),
                ("ipRouteMetric1", [("RFC1213-MIB", "ipRouteMetric1")]),
                ("ipRouteMetric2", [("RFC1213-MIB", "ipRouteMetric2")]),
                ("ipRouteMetric3", [("RFC1213-MIB", "ipRouteMetric3")]),
                ("ipRouteMetric4", [("RFC1213-MIB", "ipRouteMetric4")]),
                ("ipRouteNextHop", [("RFC1213-MIB", "ipRouteNextHop")]),
                ("ipRouteType", [("RFC1213-MIB", "ipRouteType")]),
                ("ipRouteProto", [("RFC1213-MIB", "ipRouteProto")]),
                ("ipRouteAge", [("RFC1213-MIB", "ipRouteAge")]),
                ("ipRouteMask", [("RFC1213-MIB", "ipRouteMask")]),
                ("egpInMsgs", [("RFC1213-MIB", "egpInMsgs")]),
                ("egpInErrors", [("RFC1213-MIB", "egpInErrors")]),
                ("egpOutMsgs", [("RFC1213-MIB", "egpOutMsgs")]),
                ("egpOutErrors", [("RFC1213-MIB", "egpOutErrors")]),
                ("egpNeighTable", [("RFC1213-MIB", "egpNeighTable")]),
                ("egpNeighEntry", [("RFC1213-MIB", "egpNeighEntry")]),
                ("egpNeighState", [("RFC1213-MIB", "egpNeighState")]),
                ("egpNeighAddr", [("RFC1213-MIB", "egpNeighAddr")]),
                ("egpNeighAs", [("RFC1213-MIB", "egpNeighAs")]),
                ("egpNeighInMsgs", [("RFC1213-MIB", "egpNeighInMsgs")]),
                ("egpNeighInErrs", [("RFC1213-MIB", "egpNeighInErrs")]),
                ("egpNeighOutMsgs", [("RFC1213-MIB", "egpNeighOutMsgs")]),
                ("egpNeighOutErrs", [("RFC1213-MIB", "egpNeighOutErrs")]),
                ("egpNeighInErrMsgs", [("RFC1213-MIB", "egpNeighInErrMsgs")]),
                ("egpNeighOutErrMsgs", [("RFC1213-MIB", "egpNeighOutErrMsgs")]),
                ("egpNeighStateUps", [("RFC1213-MIB", "egpNeighStateUps")]),
                ("egpNeighStateDowns", [("RFC1213-MIB", "egpNeighStateDowns")]),
                ("egpNeighIntervalHello", [("RFC1213-MIB", "egpNeighIntervalHello")]),
                ("egpNeighIntervalPoll", [("RFC1213-MIB", "egpNeighIntervalPoll")]),
                ("egpNeighMode", [("RFC1213-MIB", "egpNeighMode")]),
                ("egpNeighEventTrigger", [("RFC1213-MIB", "egpNeighEventTrigger")]),
                ("egpAs", [("RFC1213-MIB", "egpAs")]),
                ("snmpEnableAuthTraps", [("SNMPv2-MIB", "snmpEnableAuthenTraps")]),
            ),
        ),
        "RFC-1212": {"OBJECT-TYPE": [("SNMPv2-SMI", "OBJECT-TYPE")]},
        # XXX 'IndexSyntax': ???
        "RFC1213-MIB": update_dict(
            dict(commonSyms["RFC1158-MIB/RFC1213-MIB"]),
            (("PhysAddress", [("SNMPv2-TC", "PhysAddress")]),),
        ),
        "RFC-1215": {"TRAP-TYPE": [("SNMPv2-SMI", "TRAP-TYPE")]},
    }

    smiv1IdxTypes = ["INTEGER", "OCTET STRING", "IpAddress", "NetworkAddress"]

    # Name prefix and starting number of fake index object types, as generated
    # the fly to support SMIv1-only index types (e.g., "INDEX { INTEGER }").
    fakeIdxPrefix = "pysmiFakeCol"
    fakeIdxNumber = 1

    def gen_code(self, ast, symbolTable, **kwargs):
        raise NotImplementedError()

    def gen_index(self, mibsMap, **kwargs):
        raise NotImplementedError()

    @staticmethod
    def is_binary(s):
        return isinstance(s, str) and s and s[0] == "'" and s[-2:] in ("'b", "'B")

    @staticmethod
    def is_hex(s):
        return isinstance(s, str) and s and s[0] == "'" and s[-2:] in ("'h", "'H")

    def str2int(self, s):
        if self.is_binary(s):
            if s[1:-2]:
                return int(s[1:-2], 2)
            else:
                raise error.PySmiSemanticError("empty binary string to int conversion")

        elif self.is_hex(s):
            if s[1:-2]:
                return int(s[1:-2], 16)
            else:
                raise error.PySmiSemanticError("empty hex string to int conversion")
        else:
            return int(s)

    @staticmethod
    def trans_opers(symbol):
        """Convert an ASN.1 name to a "Pythonized" symbol.

        The resulting symbol must be usable as a name in Python code. As such,
        dashes are replaced with underscores, and reserved Python keywords are
        prefixed so as to make them usable as regular identifier names.

        Calling this function on a symbol that is already "Pythonized", has no
        effect, but should be avoided in general anyway.
        """
        if iskeyword(symbol):
            symbol = RESERVED_KEYWORDS_PREFIX + symbol
        return symbol.replace("-", "_")

    # compatibility with legacy code
    # Old to new attribute mapping
    deprecated_attributes = {
        "genCode": "gen_code",
        "genIndex": "gen_index",
        "isBinary": "is_binary",
        "isHex": "is_hex",
        "transOpers": "trans_opers",
    }

    def __getattr__(self, attr: str):
        if new_attr := self.deprecated_attributes.get(attr):
            warnings.warn(
                f"{attr} is deprecated. Please use {new_attr} instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            return getattr(self, new_attr)
        raise AttributeError(
            f"'{self.__class__.__name__}' object has no attribute '{attr}'"
        )



#
# This file is part of pysmi software.
#
# Copyright (c) 2015-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysmi/license.html
#
import logging

from pysmi import __version__
from pysmi import error

FLAG_NONE = 0x0000
FLAG_SEARCHER = 0x0001
FLAG_READER = 0x0002
FLAG_LEXER = 0x0004
FLAG_PARSER = 0x0008
FLAG_GRAMMAR = 0x0010
FLAG_CODEGEN = 0x0020
FLAG_WRITER = 0x0040
FLAG_COMPILER = 0x0080
FLAG_BORROWER = 0x0100
FLAG_ALL = 0xFFFF

FLAG_MAP = {
    "searcher": FLAG_SEARCHER,
    "reader": FLAG_READER,
    "lexer": FLAG_LEXER,
    "parser": FLAG_PARSER,
    "grammar": FLAG_GRAMMAR,
    "codegen": FLAG_CODEGEN,
    "writer": FLAG_WRITER,
    "compiler": FLAG_COMPILER,
    "borrower": FLAG_BORROWER,
    "all": FLAG_ALL,
}


class Printer:
    def __init__(self, logger=None, handler=None, formatter=None):
        if logger is None:
            logger = logging.getLogger("pysmi")

        logger.setLevel(logging.DEBUG)

        if handler is None:
            handler = logging.StreamHandler()

        if formatter is None:
            formatter = logging.Formatter("%(asctime)s %(name)s: %(message)s")

        handler.setFormatter(formatter)
        handler.setLevel(logging.DEBUG)

        logger.addHandler(handler)

        self.__logger = logger

    def __call__(self, msg):
        self.__logger.debug(msg)

    def __str__(self):
        """Return a string representation of the instance."""
        return "<python built-in logging>"

    def get_current_logger(self):
        return self.__logger


NullHandler = logging.NullHandler


class Debug:
    defaultPrinter = None
    _printer: Printer

    def __init__(self, *flags, **options):
        self._flags = FLAG_NONE
        if options.get("printer") is not None:
            self._printer = options.get("printer")  # type: ignore

        elif self.defaultPrinter is not None:
            self._printer = self.defaultPrinter

        else:
            if "loggerName" in options:
                # route our logs to parent logger
                self._printer = Printer(
                    logger=logging.getLogger(options["loggerName"]),
                    handler=NullHandler(),
                )
            else:
                self._printer = Printer()

        self(f"running pysmi version {__version__}")

        for flag in flags:
            inverse = flag and flag[0] in ("!", "~")

            if inverse:
                flag = flag[1:]

            try:
                if inverse:
                    self._flags &= ~FLAG_MAP[flag]
                else:
                    self._flags |= FLAG_MAP[flag]

            except KeyError:
                raise error.PySmiError(f"bad debug flag {flag}")

            self(f"debug category '{flag}' {'disabled' if inverse else 'enabled'}")

    def __str__(self):
        """Return a string representation of the instance."""
        return f"logger {self._printer}, flags {self._flags:x}"

    def __call__(self, msg):
        """Log a message."""
        self._printer(msg)

    def __and__(self, flag):
        """Return a bitwise and of the instance and a flag."""
        return self._flags & flag

    def __rand__(self, flag):
        """Return a bitwise and of a flag and the instance."""
        return flag & self._flags

    def get_current_printer(self):
        return self._printer

    def get_current_logger(self):
        return self._printer and self._printer.get_current_logger() or None


# This will yield false from bitwise and with a flag, and save
# on unnecessary calls
logger = 0


def set_logger(value):
    global logger
    logger = value



#
# This file is part of pysmi software.
#
# Copyright (c) 2015-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysmi/license.html
#
import importlib.machinery
import os
import py_compile
import sys
import tempfile

from pysmi import debug
from pysmi import error
from pysmi.compat import decode, encode
from pysmi.writer.base import AbstractWriter

SOURCE_SUFFIXES = importlib.machinery.SOURCE_SUFFIXES


class PyFileWriter(AbstractWriter):
    """Stores transformed MIB modules as Python files at specified location.

    User is expected to pass *PyFileWriter* class instance to
    *MibCompiler* on instantiation. The rest is internal to *MibCompiler*.
    """

    pyCompile = True
    pyOptimizationLevel = -1

    def __init__(self, path):
        """Creates an instance of *PyFileWriter* class.

        Args:
            path: writable directory to store Python modules
        """
        self._path = decode(os.path.normpath(path))

    def __str__(self):
        """Return a string representation of the instance."""
        return f'{self.__class__.__name__}{{"{self._path}"}}'

    def put_data(self, mibname, data, comments=(), dryRun=False):
        if dryRun:
            debug.logger & debug.FLAG_WRITER and debug.logger("dry run mode")
            return

        if not os.path.exists(self._path):
            try:
                os.makedirs(self._path)

            except OSError:
                raise error.PySmiWriterError(
                    f"failure creating destination directory {self._path}: {sys.exc_info()[1]}",
                    writer=self,
                )

        if comments:
            data = f"#{os.linesep}{os.linesep.join([f'# {x}' for x in comments])}{os.linesep}#{os.linesep}{data}"

        pyfile = os.path.join(self._path, decode(mibname))
        pyfile += SOURCE_SUFFIXES[0]

        tfile = None

        try:
            fd, tfile = tempfile.mkstemp(dir=self._path)
            os.write(fd, encode(data))
            os.close(fd)
            os.rename(tfile, pyfile)

        except (OSError, UnicodeEncodeError):
            exc = sys.exc_info()
            if tfile and os.access(tfile, os.F_OK):
                os.unlink(tfile)

            raise error.PySmiWriterError(
                f"failure writing file {pyfile}: {exc[1]}", file=pyfile, writer=self
            )

        debug.logger & debug.FLAG_WRITER and debug.logger(f"created file {pyfile}")

        if self.pyCompile:
            try:
                py_compile.compile(
                    pyfile, doraise=True, optimize=self.pyOptimizationLevel
                )

            except (SyntaxError, py_compile.PyCompileError):
                pass  # XXX

            except Exception:
                if pyfile and os.access(pyfile, os.F_OK):
                    os.unlink(pyfile)

                raise error.PySmiWriterError(
                    f"failure compiling {pyfile}: {sys.exc_info()[1]}",
                    file=mibname,
                    writer=self,
                )

        debug.logger & debug.FLAG_WRITER and debug.logger(f"{mibname} stored")

    def get_data(self, filename):
        return ""



#
# This file is part of pysmi software.
#
# Copyright (c) 2015-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysmi/license.html
#
# Package exception model:
# Here we subclass base Python exception overriding its constructor to
# accomodate error message string as its first parameter and an open
# set of keyword arguments that become exception object attributes.
# While exception object is bubbling up the call stack, intermediate
# exception handlers may insert their own attributes into exception
# object.
#


from pysmi.codegen.base import AbstractCodeGen
from pysmi.reader.base import AbstractReader
from pysmi.searcher.base import AbstractSearcher


class PySmiError(Exception):
    msg: str
    source: AbstractReader
    mibname: str
    searcher: AbstractSearcher
    handler: AbstractCodeGen

    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args)
        self.msg = args and args[0] or ""
        for k in kwargs:
            setattr(self, k, kwargs[k])

    def __repr__(self):
        """Return a string representation of the instance."""
        return f"{self.__class__.__name__}({', '.join([f'{k}={getattr(self, k)!r}' for k in dir(self) if k[0] != '_' and k != 'args'] )})"

    def __str__(self):
        """Return a string representation of the instance."""
        return self.msg


class PySmiLexerError(PySmiError):
    lineno = "?"

    def __str__(self):
        """Return a string representation of the instance."""
        return f"{self.msg}, line {self.lineno}"


class PySmiParserError(PySmiLexerError):
    pass


class PySmiSyntaxError(PySmiParserError):
    pass


class PySmiSearcherError(PySmiError):
    pass


class PySmiFileNotModifiedError(PySmiSearcherError):
    pass


class PySmiFileNotFoundError(PySmiSearcherError):
    pass


class PySmiReaderError(PySmiError):
    pass


class PySmiReaderFileNotModifiedError(PySmiReaderError):
    pass


class PySmiReaderFileNotFoundError(PySmiReaderError):
    pass


class PySmiCodegenError(PySmiError):
    pass


class PySmiSemanticError(PySmiCodegenError):
    pass


class PySmiWriterError(PySmiError):
    pass



#
# This file is part of pysmi software.
#
# Copyright (c) 2015-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysmi/license.html
#
# Build an internally used symbol table for each passed MIB.
#
from pysmi import config, debug, error, implicit_imports
from pysmi.codegen.base import AbstractCodeGen
from pysmi.mibinfo import MibInfo


class SymtableCodeGen(AbstractCodeGen):
    symsTable = {
        "MODULE-IDENTITY": ("ModuleIdentity",),
        "OBJECT-TYPE": ("MibScalar", "MibTable", "MibTableRow", "MibTableColumn"),
        "NOTIFICATION-TYPE": ("NotificationType",),
        "TEXTUAL-CONVENTION": ("TextualConvention",),
        "MODULE-COMPLIANCE": ("ModuleCompliance",),
        "OBJECT-GROUP": ("ObjectGroup",),
        "NOTIFICATION-GROUP": ("NotificationGroup",),
        "AGENT-CAPABILITIES": ("AgentCapabilities",),
        "OBJECT-IDENTITY": ("ObjectIdentity",),
        "TRAP-TYPE": ("NotificationType",),  # smidump always uses NotificationType
        "BITS": ("Bits",),
    }

    constImports = {
        "SNMPv2-SMI": (
            "iso",
            "Bits",  # XXX
            "Integer32",  # XXX
            "TimeTicks",  # bug in some IETF MIBs
            "Counter32",  # bug in some IETF MIBs (e.g. DSA-MIB)
            "Counter64",  # bug in some MIBs (e.g.A3COM-HUAWEI-LswINF-MIB)
            "NOTIFICATION-TYPE",  # bug in some MIBs (e.g. A3COM-HUAWEI-DHCPSNOOP-MIB)
            "Gauge32",  # bug in some IETF MIBs (e.g. DSA-MIB)
            "MODULE-IDENTITY",
            "OBJECT-TYPE",
            "OBJECT-IDENTITY",
            "Unsigned32",
            "IpAddress",  # XXX
            "MibIdentifier",
        ),  # OBJECT IDENTIFIER
        "SNMPv2-TC": (
            "DisplayString",
            "PhysAddress",
            "TEXTUAL-CONVENTION",
        ),  # XXX
        "SNMPv2-CONF": (
            "MODULE-COMPLIANCE",
            "NOTIFICATION-GROUP",
        ),  # XXX
    }

    baseTypes = ["Integer", "Integer32", "Bits", "ObjectIdentifier", "OctetString"]

    typeClasses = {
        "COUNTER32": "Counter32",
        "COUNTER64": "Counter64",
        "GAUGE32": "Gauge32",
        "INTEGER": "Integer32",  # XXX
        "INTEGER32": "Integer32",
        "IPADDRESS": "IpAddress",
        "NETWORKADDRESS": "IpAddress",
        "OBJECT IDENTIFIER": "ObjectIdentifier",
        "OCTET STRING": "OctetString",
        "OPAQUE": "Opaque",
        "TIMETICKS": "TimeTicks",
        "UNSIGNED32": "Unsigned32",
        "Counter": "Counter32",
        "Gauge": "Gauge32",
        "NetworkAddress": "IpAddress",  # RFC1065-SMI, RFC1155-SMI -> SNMPv2-SMI
        "nullSpecific": "zeroDotZero",  # RFC1158-MIB -> SNMPv2-SMI
        "ipRoutingTable": "ipRouteTable",  # RFC1158-MIB -> RFC1213-MIB
        "snmpEnableAuthTraps": "snmpEnableAuthenTraps",  # RFC1158-MIB -> SNMPv2-MIB
    }

    def __init__(self):
        self._rows = set()  # symbols
        self._cols = {}  # k, v = name, datatype [name is *not* a Pythonized symbol!]
        self._sequenceTypes = set()
        self._exports = set()
        self._postponedSyms = {}  # k, v = symbol, (parents, properties)
        self._parentOids = set()
        self._importMap = {}  # k, v = symbol, MIB
        self._symsOrder = []
        self._out = {}  # k, v = symbol, properties
        self.moduleName = ["DUMMY"]
        self._moduleRevision = None
        self.genRules = {"text": True}

    def sym_trans(self, symbol):
        if symbol in self.symsTable:
            return self.symsTable[symbol]

        return (symbol,)

    def prep_data(self, pdata, classmode=False):
        data = []
        for el in pdata:
            if not isinstance(el, tuple):
                data.append(el)

            elif len(el) == 1:
                data.append(el[0])

            else:
                data.append(
                    self.handlersTable[el[0]](
                        self,
                        self.prep_data(el[1:], classmode=classmode),
                        classmode=classmode,
                    )
                )

        return data

    def gen_imports(self, imports, apply_implicit=False):
        # normalize incoming imports into a mutable mapping of lists
        if not isinstance(imports, dict):
            try:
                imports = dict(imports)
            except Exception:
                imports = {}

        # ensure all import value containers are lists we can append to
        for m in list(imports):
            imports[m] = list(imports[m])

        # helper to safely add a symbol to imports[module]
        def _add_import(mod, sym):
            if mod in imports:
                if isinstance(imports[mod], list):
                    imports[mod].append(sym)
                else:
                    try:
                        imports[mod] = list(imports[mod]) + [sym]
                    except Exception:
                        imports[mod] = [sym]
            else:
                imports[mod] = [sym]

        # convertion to SNMPv2
        toDel = []
        for module in list(imports):
            if module in self.convertImportv2:
                for symbol in imports[module]:
                    if symbol in self.convertImportv2[module]:
                        toDel.append((module, symbol))

                        for newImport in self.convertImportv2[module][symbol]:
                            newModule, newSymbol = newImport
                            _add_import(newModule, newSymbol)

        # removing converted symbols
        for d in toDel:
            imports[d[0]].remove(d[1])

        # merging mib and constant imports (ensure lists)
        for module in self.constImports:
            for sym in self.constImports[module]:
                _add_import(module, sym)

        # apply per-MIB implicit imports only when explicitly requested
        if apply_implicit:
            try:
                mib_exceptions = implicit_imports.IMPLICIT_IMPORTS.get(
                    self.moduleName[0]
                )
            except Exception:
                mib_exceptions = None

            if mib_exceptions:
                for newModule, newSymbol in mib_exceptions:
                    _add_import(newModule, newSymbol)

        for module in sorted(imports):
            symbols = ()
            for symbol in set(imports[module]):
                symbols += self.sym_trans(symbol)

            if symbols:
                self._importMap.update([(self.trans_opers(s), module) for s in symbols])

        return {}, tuple(sorted(imports))

    def all_parents_exists(self, parents):
        parentsExists = True
        for parent in parents:
            if not (
                parent in self._out
                or parent in self._importMap
                or parent in self.baseTypes
                or parent in ("MibTable", "MibTableRow", "MibTableColumn")
                or parent in self._rows
            ):
                parentsExists = False
                break

        return parentsExists

    def reg_sym(self, symbol, symProps, parents=()):
        if (
            symbol in self._out
            or symbol in self._postponedSyms
            or (config.STRICT_MODE and symbol in self._importMap)
        ):
            raise error.PySmiSemanticError(f"Duplicate symbol found: {symbol}")

        if self.all_parents_exists(parents):
            self._out[symbol] = symProps
            self._symsOrder.append(symbol)
            self.reg_postponed_syms()

        else:
            self._postponedSyms[symbol] = (parents, symProps)

    def reg_postponed_syms(self):
        regedSyms = []
        for sym, val in self._postponedSyms.items():
            parents, symProps = val

            if self.all_parents_exists(parents):
                self._out[sym] = symProps
                self._symsOrder.append(sym)
                regedSyms.append(sym)

        for sym in regedSyms:
            self._postponedSyms.pop(sym)

        # Clause handlers

    # noinspection PyUnusedLocal
    def gen_agent_capabilities(self, data, classmode=False):
        origName, release, status, description, reference, oid = data

        pysmiName = self.trans_opers(origName)

        symProps = {"type": "AgentCapabilities", "oid": oid, "origName": origName}

        self.reg_sym(pysmiName, symProps)

    # noinspection PyUnusedLocal
    def gen_module_identity(self, data, classmode=False):
        (
            origName,
            lastUpdated,
            organization,
            contactInfo,
            description,
            revisions,
            oid,
        ) = data

        pysmiName = self.trans_opers(origName)

        symProps = {"type": "ModuleIdentity", "oid": oid, "origName": origName}

        if revisions:
            self._moduleRevision = revisions[0]

        self.reg_sym(pysmiName, symProps)

    # noinspection PyUnusedLocal
    def gen_module_compliance(self, data, classmode=False):
        origName, status, description, reference, compliances, oid = data

        pysmiName = self.trans_opers(origName)

        symProps = {"type": "ModuleCompliance", "oid": oid, "origName": origName}

        self.reg_sym(pysmiName, symProps)

    # noinspection PyUnusedLocal
    def gen_notification_group(self, data, classmode=False):
        origName, objects, status, description, reference, oid = data

        pysmiName = self.trans_opers(origName)

        symProps = {"type": "NotificationGroup", "oid": oid, "origName": origName}

        self.reg_sym(pysmiName, symProps)

    # noinspection PyUnusedLocal
    def gen_notification_type(self, data, classmode=False):
        origName, objects, status, description, reference, oid = data

        pysmiName = self.trans_opers(origName)

        symProps = {"type": "NotificationType", "oid": oid, "origName": origName}

        self.reg_sym(pysmiName, symProps)

    # noinspection PyUnusedLocal
    def gen_object_group(self, data, classmode=False):
        origName, objects, status, description, reference, oid = data

        pysmiName = self.trans_opers(origName)

        symProps = {"type": "ObjectGroup", "oid": oid, "origName": origName}

        self.reg_sym(pysmiName, symProps)

    # noinspection PyUnusedLocal
    def gen_object_identity(self, data, classmode=False):
        origName, status, description, reference, oid = data

        pysmiName = self.trans_opers(origName)

        symProps = {"type": "ObjectIdentity", "oid": oid, "origName": origName}

        self.reg_sym(pysmiName, symProps)

    # noinspection PyUnusedLocal
    def gen_object_type(self, data, classmode=False):
        (
            origName,
            syntax,
            units,
            maxaccess,
            status,
            description,
            reference,
            augmention,
            index,
            defval,
            oid,
        ) = data

        pysmiName = self.trans_opers(origName)

        symProps = {
            "type": "ObjectType",
            "oid": oid,
            "syntax": syntax,  # (type, module), subtype
            "origName": origName,
        }

        parents = [syntax[0][0]]

        if augmention:
            parents.append(self.trans_opers(augmention))

        if index and index[1]:
            namepart, fakeIndexes, fakeSymSyntax = index
            for fakeIdx, fakeSyntax in zip(fakeIndexes, fakeSymSyntax):
                fakeName = namepart + str(fakeIdx)

                fakeSymProps = {
                    "type": "fakeColumn",
                    "oid": oid + (fakeIdx,),
                    "syntax": fakeSyntax,
                    "origName": fakeName,
                }

                self.reg_sym(fakeName, fakeSymProps)

        self.reg_sym(pysmiName, symProps, parents)

    # noinspection PyUnusedLocal
    def gen_trap_type(self, data, classmode=False):
        origName, enterprise, variables, description, reference, value = data

        pysmiName = self.trans_opers(origName)

        symProps = {
            "type": "NotificationType",
            "oid": enterprise + (0, value),
            "origName": origName,
        }

        self.reg_sym(pysmiName, symProps)

    # noinspection PyUnusedLocal
    def gen_type_declaration(self, data, classmode=False):
        origName, declaration = data

        pysmiName = self.trans_opers(origName)

        if declaration:
            parentType, attrs, isTC = declaration

            if parentType:  # skipping SEQUENCE case
                symProps = {
                    "type": "TypeDeclaration",
                    "syntax": (parentType, attrs),  # (type, module), subtype
                    "origName": origName,
                    "isTC": isTC,
                }

                self.reg_sym(pysmiName, symProps, [declaration[0][0]])
            else:
                self._sequenceTypes.add(pysmiName)

    # noinspection PyUnusedLocal
    def gen_value_declaration(self, data, classmode=False):
        origName, oid = data

        pysmiName = self.trans_opers(origName)

        symProps = {"type": "MibIdentifier", "oid": oid, "origName": origName}

        self.reg_sym(pysmiName, symProps)

    # Subparts generation functions
    # noinspection PyUnusedLocal,PyMethodMayBeStatic
    def gen_bit_names(self, data, classmode=False):
        names = data[0]
        return names

    # noinspection PyUnusedLocal,PyMethodMayBeStatic
    def gen_bits(self, data, classmode=False):
        bits = dict(data[0])
        return ("Bits", ""), bits

    # noinspection PyUnusedLocal,PyUnusedLocal,PyMethodMayBeStatic
    def gen_compliances(self, data, classmode=False):
        return ""

    # noinspection PyUnusedLocal
    def gen_conceptual_table(self, data, classmode=False):
        row = data[0]
        if row[0] and row[0][0]:
            self._rows.add(row[0][0])  # (already a Pythonized symbol)
        return ("MibTable", ""), ""

    # noinspection PyUnusedLocal,PyUnusedLocal,PyMethodMayBeStatic
    def gen_contact_info(self, data, classmode=False):
        return ""

    # noinspection PyUnusedLocal,PyUnusedLocal,PyMethodMayBeStatic
    def gen_display_hint(self, data, classmode=False):
        return ""

    # noinspection PyUnusedLocal,PyUnusedLocal,PyMethodMayBeStatic
    def gen_def_val(self, data, classmode=False):
        return ""

    # noinspection PyUnusedLocal,PyUnusedLocal,PyMethodMayBeStatic
    def gen_description(self, data, classmode=False):
        return ""

    def gen_reference(self, data, classmode=False):
        return ""

    def gen_status(self, data, classmode=False):
        return ""

    def gen_product_release(self, data, classmode=False):
        return ""

    def gen_enum_spec(self, data, classmode=False):
        return self.gen_bits(data, classmode=classmode)[1]

    def gen_index(self, data, classmode=False):
        indexes = data[0]

        fakeIndexes, fakeSymsSyntax = [], []

        for idx in indexes:
            idxName = idx[1]
            if idxName in self.smiv1IdxTypes:  # SMIv1 support
                idxType = idxName

                objType = self.typeClasses.get(idxType, idxType)
                objType = self.trans_opers(objType)

                fakeIndexes.append(self.fakeIdxNumber)
                fakeSymsSyntax.append((("MibTableColumn", ""), objType))
                self.fakeIdxNumber += 1

        return self.fakeIdxPrefix, fakeIndexes, fakeSymsSyntax

    # noinspection PyUnusedLocal,PyUnusedLocal,PyMethodMayBeStatic
    def gen_integer_subtype(self, data, classmode=False):
        return [data[0]]

    # noinspection PyUnusedLocal,PyUnusedLocal,PyMethodMayBeStatic
    def gen_max_access(self, data, classmode=False):
        return ""

    # noinspection PyUnusedLocal,PyUnusedLocal,PyMethodMayBeStatic
    def gen_octetstring_subtype(self, data, classmode=False):
        return [data[0]]

    # noinspection PyUnusedLocal
    def gen_oid(self, data, classmode=False):
        out = ()
        for el in data[0]:
            if isinstance(el, str):
                parent = self.trans_opers(el)
                self._parentOids.add(parent)
                out += ((parent, self._importMap.get(parent, self.moduleName[0])),)

            elif isinstance(el, int):
                out += (el,)

            elif isinstance(el, tuple):
                out += (el[1],)  # XXX Do we need to create a new object el[0]?

            else:
                raise error.PySmiSemanticError(f"unknown datatype for OID: {el}")

        return out

    # noinspection PyUnusedLocal,PyUnusedLocal,PyMethodMayBeStatic
    def gen_objects(self, data, classmode=False):
        return ""

    # noinspection PyUnusedLocal,PyUnusedLocal,PyMethodMayBeStatic
    def gen_time(self, data, classmode=False):
        return ""

    # noinspection PyUnusedLocal,PyUnusedLocal,PyMethodMayBeStatic
    def gen_last_updated(self, data, classmode=False):
        return data[0]

    # noinspection PyUnusedLocal,PyUnusedLocal,PyMethodMayBeStatic
    def gen_organization(self, data, classmode=False):
        return data[0]

    # noinspection PyUnusedLocal,PyUnusedLocal,PyMethodMayBeStatic
    def gen_revisions(self, data, classmode=False):
        lastRevision, lastDescription = data[0][0][0], data[0][0][1][1]
        return lastRevision, lastDescription

    def gen_row(self, data, classmode=False):
        row = data[0]
        row = self.trans_opers(row)
        return (
            row in self._rows
            and (("MibTableRow", ""), "")
            or self.gen_simple_syntax(data, classmode=classmode)
        )

    # noinspection PyUnusedLocal
    def gen_sequence(self, data, classmode=False):
        cols = data[0]
        self._cols.update(cols)
        return "", ""

    # noinspection PyUnusedLocal
    def gen_simple_syntax(self, data, classmode=False):
        objType = data[0]

        module = ""

        objType = self.typeClasses.get(objType, objType)
        objType = self.trans_opers(objType)

        if objType not in self.baseTypes:
            module = self._importMap.get(objType, self.moduleName[0])

        subtype = len(data) == 2 and data[1] or ""

        return (objType, module), subtype

    # noinspection PyUnusedLocal,PyMethodMayBeStatic
    def gen_type_declaration_rhs(self, data, classmode=False):
        if len(data) == 1:
            parentType, attrs = data[0]  # just syntax
            isTC = False

        else:
            # Textual convention
            display, status, description, reference, syntax = data
            parentType, attrs = syntax
            isTC = True

        return parentType, attrs, isTC

    # noinspection PyUnusedLocal,PyUnusedLocal,PyMethodMayBeStatic
    def gen_units(self, data, classmode=False):
        return ""

    handlersTable = {
        "agentCapabilitiesClause": gen_agent_capabilities,
        "moduleIdentityClause": gen_module_identity,
        "moduleComplianceClause": gen_module_compliance,
        "notificationGroupClause": gen_notification_group,
        "notificationTypeClause": gen_notification_type,
        "objectGroupClause": gen_object_group,
        "objectIdentityClause": gen_object_identity,
        "objectTypeClause": gen_object_type,
        "trapTypeClause": gen_trap_type,
        "typeDeclaration": gen_type_declaration,
        "valueDeclaration": gen_value_declaration,
        "ApplicationSyntax": gen_simple_syntax,
        "BitNames": gen_bit_names,
        "BITS": gen_bits,
        "ComplianceModules": gen_compliances,
        "conceptualTable": gen_conceptual_table,
        "CONTACT-INFO": gen_contact_info,
        "DISPLAY-HINT": gen_display_hint,
        "DEFVAL": gen_def_val,
        "DESCRIPTION": gen_description,
        "REFERENCE": gen_reference,
        "Status": gen_status,
        "PRODUCT-RELEASE": gen_product_release,
        "enumSpec": gen_enum_spec,
        "INDEX": gen_index,
        "integerSubType": gen_integer_subtype,
        "MaxAccessPart": gen_max_access,
        "Notifications": gen_objects,
        "octetStringSubType": gen_octetstring_subtype,
        "objectIdentifier": gen_oid,
        "Objects": gen_objects,
        "LAST-UPDATED": gen_last_updated,
        "ORGANIZATION": gen_organization,
        "Revisions": gen_revisions,
        "row": gen_row,
        "SEQUENCE": gen_sequence,
        "SimpleSyntax": gen_simple_syntax,
        "typeDeclarationRHS": gen_type_declaration_rhs,
        "UNITS": gen_units,
        "VarTypes": gen_objects,
    }

    def correct_postponed_syms(self):
        """
        We are about to fail on unresolvable symbols.

        See if we can correct any cases to make them valid after all.
        """
        for sym, val in self._postponedSyms.items():
            parents, symProps = val

            if symProps["type"] == "ObjectType":
                syntax = symProps["syntax"]

                # Leniency for non-conforming MIBs: if the object type is not
                # known as a table row, but it has a SEQUENCE-type syntax, then
                # turn it into a table row after all. This leniency exception
                # allows the "SEQUENCE OF" type name of the conceptual table to
                # contain typos in common cases.
                if syntax[0][0] in self._sequenceTypes:
                    self._rows.add(syntax[0][0])
                    symProps["syntax"] = (("MibTableRow", ""), "")
                    parents[0] = "MibTableRow"

        self.reg_postponed_syms()

    def gen_code(self, ast, symbolTable, **kwargs):
        self.genRules["text"] = kwargs.get("genTexts", False)
        self._rows.clear()
        self._cols.clear()
        self._parentOids.clear()
        self._symsOrder = []
        self._postponedSyms.clear()
        self._importMap.clear()
        self._out = {}  # should be new object, do not use `clear` method
        self.moduleName[0], moduleOid, imports, declarations = ast
        # First attempt without implicit imports
        try:
            out, importedModules = self.gen_imports(imports, apply_implicit=False)

            for declr in declarations or []:
                if declr:
                    clausetype = declr[0]
                    classmode = clausetype == "typeDeclaration"
                    self.handlersTable[declr[0]](
                        self, self.prep_data(declr[1:], classmode), classmode
                    )

            if self._postponedSyms:
                self.correct_postponed_syms()
                if self._postponedSyms:
                    raise error.PySmiSemanticError(
                        f"Unknown parents for symbols: {', '.join(self._postponedSyms)}"
                    )

            for sym in self._parentOids:
                if sym not in self._out and sym not in self._importMap:
                    raise error.PySmiSemanticError(f"Unknown parent symbol: {sym}")

        except error.PySmiSemanticError:
            # Try again only if we have implicit imports defined for this MIB
            if implicit_imports.IMPLICIT_IMPORTS.get(self.moduleName[0]):
                # reset state and retry with implicit imports applied
                self._rows.clear()
                self._cols.clear()
                self._parentOids.clear()
                self._symsOrder = []
                self._postponedSyms.clear()
                self._importMap.clear()
                self._out = {}

                out, importedModules = self.gen_imports(imports, apply_implicit=True)

                for declr in declarations or []:
                    if declr:
                        clausetype = declr[0]
                        classmode = clausetype == "typeDeclaration"
                        self.handlersTable[declr[0]](
                            self, self.prep_data(declr[1:], classmode), classmode
                        )

                if self._postponedSyms:
                    self.correct_postponed_syms()
                    if self._postponedSyms:
                        raise error.PySmiSemanticError(
                            f"Unknown parents for symbols: {', '.join(self._postponedSyms)}"
                        )

                for sym in self._parentOids:
                    if sym not in self._out and sym not in self._importMap:
                        raise error.PySmiSemanticError(f"Unknown parent symbol: {sym}")
            else:
                raise

        self._out["_symtable_order"] = list(self._symsOrder)
        self._out["_symtable_cols"] = list(self._cols)
        self._out["_symtable_rows"] = list(self._rows)

        debug.logger & debug.FLAG_CODEGEN and debug.logger(
            f"canonical MIB name {self.moduleName[0]} ({moduleOid}), imported MIB(s) {','.join(importedModules) or '<none>'}, Symbol table size {len(self._out)} symbols"
        )

        return (
            MibInfo(
                oid=None,
                name=self.moduleName[0],
                revision=self._moduleRevision,
                imported=tuple(x for x in importedModules),
            ),
            self._out,
        )



#
# This file is part of pysmi software.
#
# Copyright (c) 2015-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysmi/license.html
#
import importlib.machinery
import importlib.util
import os
import struct
import sys
import time

from pysmi import debug
from pysmi import error
from pysmi.compat import decode
from pysmi.searcher.base import AbstractSearcher

PY_MAGIC_NUMBER = importlib.util.MAGIC_NUMBER
SOURCE_SUFFIXES = importlib.machinery.SOURCE_SUFFIXES
BYTECODE_SUFFIXES = importlib.machinery.BYTECODE_SUFFIXES


class PyFileSearcher(AbstractSearcher):
    """Figures out if given Python file (source or bytecode) exists at given location."""

    def __init__(self, path):
        """Create an instance of *PyFileSearcher* bound to specific directory.

        Args:
          path (str): path to local directory
        """
        self._path = os.path.normpath(decode(path))

    def __str__(self):
        """Return a string representation of the instance."""
        return f'{self.__class__.__name__}{{"{self._path}"}}'

    def file_exists(self, mibname, mtime, rebuild=False):
        if rebuild:
            debug.logger & debug.FLAG_SEARCHER and debug.logger(
                f"pretend {mibname} is very old"
            )
            return

        mibname = decode(mibname)
        pyfile = os.path.join(self._path, mibname)

        for pySfx in BYTECODE_SUFFIXES:
            f = pyfile + pySfx

            if not os.path.exists(f) or not os.path.isfile(f):
                debug.logger & debug.FLAG_SEARCHER and debug.logger(
                    f"{f} not present or not a file"
                )
                continue

            try:
                fp = open(f, "rb")
                pyData = fp.read(8)
                fp.close()

            except OSError:
                raise error.PySmiSearcherError(
                    f"failure opening compiled file {f}: {sys.exc_info()[1]}",
                    searcher=self,
                )
            if pyData[:4] == PY_MAGIC_NUMBER:
                pyData = pyData[4:]
                pyTime = struct.unpack("<L", pyData[:4])[0]
                debug.logger & debug.FLAG_SEARCHER and debug.logger(
                    f"found {f}, mtime {time.strftime('%a, %d %b %Y %H:%M:%S GMT', time.gmtime(pyTime))}"
                )
                if pyTime >= mtime:
                    raise error.PySmiFileNotModifiedError()

                else:
                    raise error.PySmiFileNotFoundError(
                        f"older file {mibname} exists", searcher=self
                    )

            else:
                debug.logger & debug.FLAG_SEARCHER and debug.logger(f"bad magic in {f}")
                continue

        for pySfx in SOURCE_SUFFIXES:
            f = pyfile + pySfx

            if not os.path.exists(f) or not os.path.isfile(f):
                debug.logger & debug.FLAG_SEARCHER and debug.logger(
                    f"{f} not present or not a file"
                )
                continue

            try:
                pyTime = os.stat(f)[8]

            except OSError:
                raise error.PySmiSearcherError(
                    f"failure opening compiled file {f}: {sys.exc_info()[1]}",
                    searcher=self,
                )

            debug.logger & debug.FLAG_SEARCHER and debug.logger(
                f"found {f}, mtime {time.strftime('%a, %d %b %Y %H:%M:%S GMT', time.gmtime(pyTime))}"
            )

            if pyTime >= mtime:
                raise error.PySmiFileNotModifiedError()

        raise error.PySmiFileNotFoundError(
            f"no compiled file {mibname} found", searcher=self
        )



#
# This file is part of pysmi software.
#
# Copyright (c) 2015-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysmi/license.html
#
import time

from pysmi import debug, error
from pysmi.mibinfo import MibInfo
from pysmi.reader.base import AbstractReader


class CallbackReader(AbstractReader):
    """Fetch ASN.1 MIB text by name by calling user-defined callable.

    *CallbackReader* class instance tries to retrieve ASN.1 MIB files
    by name and return their contents to caller.
    """

    def __init__(self, cbFun, cbCtx=None):
        """Create an instance of *CallbackReader* bound to specific URL.

        Args:
            cbFun (callable): user callable accepting *MIB name* and *cbCtx* objects

        Keyword Args:
            cbCtx (object): user object that can be used to communicate state information
                between user-scope code and the *cbFun* callable scope
        """
        self._cbFun = cbFun
        self._cbCtx = cbCtx

    def __str__(self):
        """Return string representation of the instance."""
        return f'{self.__class__.__name__}{{"{self._cbFun}"}}'

    def get_data(self, mibname, **options):
        debug.logger & debug.FLAG_READER and debug.logger(
            f"calling user callback {self._cbFun} for MIB {mibname}"
        )

        res = self._cbFun(mibname, self._cbCtx)
        if res:
            return (
                MibInfo(
                    path="file:///dev/stdin", file="", name=mibname, mtime=time.time()
                ),
                res,
            )

        raise error.PySmiReaderFileNotFoundError(mibname=mibname, reader=self)



#
# This file is part of pysmi software.
#
# Copyright (c) 2015-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysmi/license.html
#


class AbstractWriter:
    def set_options(self, **kwargs):
        for k in kwargs:
            setattr(self, k, kwargs[k])
        return self

    def put_data(self, mibname, data, comments=(), dryRun=False):
        raise NotImplementedError()

    def get_data(self, filename):
        raise NotImplementedError()



"""Per-MIB implicit imports workarounds.

Populate this mapping only when necessary. Keep entries minimal and
document why each entry exists (vendor, MIB, date).
"""

from typing import Dict, List, Tuple

IMPLICIT_IMPORTS: Dict[str, List[Tuple[str, str]]] = {
    # 'JUNIPER-SMI': [('SNMPv2-SMI', 'Opaque')],  ## Example entry. Its latest revision missed importing Opaque for Integer64 textual convention.
}



"""
Compile SMIv1/v2 MIBs
+++++++++++++++++++++

Look up specific ASN.1 MIBs at configured local directories,
compile them into pysnmp form if not done yet and save Python
modules as plain-text files in a local directory.

Try to support both SMIv1 and SMIv2 flavors of SMI as well as
popular deviations from official syntax found in the wild.

When figuring out if compilation is needed, check all known
places where pysnmp MIBs could possibly be found.

You can force MIB re-compilation by passing rebuild flag to
MIB compiler (see below).

Default invocation of MIB compiler does not generate [potentially
large] comments and texts found in MIBs. If you need them in pysnmp
MIB modules, just pass genTexts flag to MIB compiler.
"""  #
from pysmi.reader import FileReader
from pysmi.searcher import PyFileSearcher, PyPackageSearcher, StubSearcher
from pysmi.writer import PyFileWriter
from pysmi.parser import SmiStarParser
from pysmi.codegen import PySnmpCodeGen
from pysmi.compiler import MibCompiler

inputMibs = ["IF-MIB", "IP-MIB"]
srcDirectories = ["/usr/share/snmp/mibs"]
dstDirectory = ".pysnmp-mibs"

# Initialize compiler infrastructure

mibCompiler = MibCompiler(SmiStarParser(), PySnmpCodeGen(), PyFileWriter(dstDirectory))

# search for source MIBs here
mibCompiler.add_sources(*[FileReader(x) for x in srcDirectories])

# check compiled MIBs in our own productions
mibCompiler.add_searchers(PyFileSearcher(dstDirectory))
# ...and at default PySNMP MIBs packages
mibCompiler.add_searchers(
    *[PyPackageSearcher(x) for x in PySnmpCodeGen.defaultMibPackages]
)

# never recompile MIBs with MACROs
mibCompiler.add_searchers(StubSearcher(*PySnmpCodeGen.baseMibs))

# run [possibly recursive] MIB compilation
results = mibCompiler.compile(*inputMibs)  # , rebuild=True, genTexts=True)

print(f"Results: {', '.join(f'{x}:{results[x]}' for x in results)}")



#
# This file is part of pysmi software.
#
# Copyright (c) 2015-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysmi/license.html
#
import sys


def encode(s):
    if isinstance(s, str):
        s = s.encode("utf-8", "ignore")
    return s


def decode(s):
    if isinstance(s, bytes):
        s = s.decode("utf-8", "ignore")
    return s



#
# This file is part of pysmi software.
#
# Copyright (c) 2015-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysmi/license.html
#
import os


class AbstractReader:
    maxMibSize = 10000000  # MIBs can't be that large
    fuzzyMatching = True  # try different file names while searching for MIB
    originalMatching = uppercaseMatching = lowcaseMatching = True
    exts = ["", os.path.extsep + "txt", os.path.extsep + "mib", os.path.extsep + "my"]
    exts.extend([x.upper() for x in exts if x])

    def set_options(self, **kwargs):
        for k in kwargs:
            setattr(self, k, kwargs[k])
        return self

    def get_mib_variants(self, mibname, **options):
        filenames = []

        if self.originalMatching:
            filenames.append(mibname)

        if self.uppercaseMatching:
            filenames.append(mibname.upper())

        if self.lowcaseMatching:
            filenames.append(mibname.lower())

        if self.fuzzyMatching:
            part = filenames[-1].find("-mib")
            if part != -1:
                filenames.extend([x[:part] for x in filenames])
            else:
                suffixed = mibname + "-mib"
                filenames.append(suffixed.upper())
                filenames.append(suffixed.lower())

        return ((x, x + y) for x in filenames for y in options.get("exts", self.exts))

    def get_data(self, filename, **options):
        raise NotImplementedError()



#
# This file is part of pysmi software.
#
# Copyright (c) 2015-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysmi/license.html
#
import warnings
from urllib import parse as urlparse
from urllib.request import url2pathname

from pysmi import error
from pysmi.reader.httpclient import HttpReader
from pysmi.reader.localfile import FileReader
from pysmi.reader.zipreader import ZipReader


def get_readers_from_urls(*sourceUrls, **options):
    readers = []
    for sourceUrl in sourceUrls:
        mibSource = urlparse.urlparse(sourceUrl)

        if mibSource.scheme in ("", "file", "zip"):
            scheme = mibSource.scheme
            filePath = url2pathname(mibSource.path)
            if scheme != "file" and (
                filePath.endswith(".zip") or filePath.endswith(".ZIP")
            ):
                scheme = "zip"

            else:
                scheme = "file"

            if scheme == "file":
                readers.append(FileReader(filePath).set_options(**options))
            else:
                readers.append(ZipReader(filePath).set_options(**options))

        elif mibSource.scheme in ("http", "https"):
            readers.append(HttpReader(sourceUrl).set_options(**options))

        else:
            raise error.PySmiError(f"Unsupported URL scheme {sourceUrl}")

    return readers


# Compatibility API
deprecated_attributes = {
    "getReadersFromUrls": "get_readers_from_urls",
}


def __getattr__(attr: str):
    """Handle deprecated attributes."""
    if newAttr := deprecated_attributes.get(attr):
        warnings.warn(
            f"{attr} is deprecated. Please use {newAttr} instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return globals()[newAttr]
    raise AttributeError(f"module {__name__} has no attribute {attr}")



STRICT_MODE: bool = False
"""
A boolean flag indicating whether strict mode is enabled or not.

In strict mode, the configuration enforces strict validation rules and raises exceptions for any violations.
When strict mode is disabled, the configuration allows more flexibility and may tolerate certain violations.
"""



"""
Always borrow pysnmp MIBs
+++++++++++++++++++++++++

Try to borrow precompiled pysnmp MIB file(s) from a web-site.

In this example no attempt is made to find and compile ASN.1
MIB source.

Fetched pysnmp MIB(s) are stored in a local directory.
"""  #
from pysmi.reader import HttpReader
from pysmi.searcher import PyFileSearcher
from pysmi.borrower import PyFileBorrower
from pysmi.writer import PyFileWriter
from pysmi.parser import NullParser
from pysmi.codegen import NullCodeGen
from pysmi.compiler import MibCompiler

inputMibs = ["MIKROTIK-MIB"]


httpBorrowers = [("mibs.pysnmp.com", 443, "/pysnmp/notexts/@mib@")]
dstDirectory = ".pysnmp-mibs"

# Initialize compiler infrastructure

mibCompiler = MibCompiler(NullParser(), NullCodeGen(), PyFileWriter(dstDirectory))

# check compiled/borrowed MIBs in our own productions
mibCompiler.add_searchers(PyFileSearcher(dstDirectory))

# search for precompiled MIBs at Web sites
mibCompiler.add_borrowers(
    *[PyFileBorrower(HttpReader("https://mibs.pysnmp.com/notexts/@mib@"))]
)

# run MIB compilation
results = mibCompiler.compile(*inputMibs)

print(f"Results: {', '.join(f'{x}:{results[x]}' for x in results)}")



#
# This file is part of pysmi software.
#
# Copyright (c) 2015-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysmi/license.html
#
import os
import sys
import time

from pysmi import debug
from pysmi import error
from pysmi.compat import decode
from pysmi.mibinfo import MibInfo
from pysmi.reader.base import AbstractReader


class FileReader(AbstractReader):
    """Fetch ASN.1 MIB text by name from local file.

    *FileReader* class instance tries to locate ASN.1 MIB files
    by name, fetch and return their contents to caller.
    """

    useIndexFile = True  # optional .index file mapping MIB to file name
    indexFile = ".index"

    def __init__(self, path, recursive=True, ignoreErrors=True):
        """Create an instance of *FileReader* serving a directory.

        Args:
            path (str): directory to search MIB files

        Keyword Args:
            recursive (bool): whether to include subdirectories
            ignoreErrors (bool): ignore filesystem access errors
        """
        self._path = os.path.normpath(path)
        self._recursive = recursive
        self._ignoreErrors = ignoreErrors
        self._indexLoaded = False
        self._mibIndex = None

    def __str__(self):
        """Return string representation of the instance."""
        return f'{self.__class__.__name__}{{"{self._path}"}}'

    def get_subdirs(self, path, recursive=True, ignoreErrors=True):
        if not recursive:
            return [path]

        dirs = [path]

        try:
            subdirs = os.listdir(path)

        except OSError:
            if ignoreErrors:
                return dirs

            else:
                raise error.PySmiError(
                    f"directory {path} access error: {sys.exc_info()[1]}"
                )

        for d in subdirs:
            d = os.path.join(decode(path), decode(d))
            if os.path.isdir(d):
                dirs.extend(self.get_subdirs(d, recursive))

        return dirs

    @staticmethod
    def load_index(indexFile):
        mibIndex = {}
        if os.path.exists(indexFile):
            try:
                f = open(indexFile)
                mibIndex = dict([x.split()[:2] for x in f.readlines()])
                f.close()
                debug.logger & debug.FLAG_READER and debug.logger(
                    f"loaded MIB index map from {indexFile} file, {len(mibIndex)} entries"
                )

            except OSError:
                pass

        return mibIndex

    def get_mib_variants(self, mibname, **options):
        if self.useIndexFile:
            if not self._indexLoaded:
                self._mibIndex = self.load_index(
                    os.path.join(self._path, self.indexFile)
                )
                self._indexLoaded = True

            if mibname in self._mibIndex:
                debug.logger & debug.FLAG_READER and debug.logger(
                    f"found {mibname} in MIB index: {self._mibIndex[mibname]}"
                )
                return [(mibname, self._mibIndex[mibname])]

        return super().get_mib_variants(mibname, **options)

    def get_data(self, mibname, **options):
        debug.logger & debug.FLAG_READER and debug.logger(
            f"looking for MIB {mibname}{' recursively' if self._recursive else ''}"
        )

        for path in self.get_subdirs(self._path, self._recursive, self._ignoreErrors):
            for mibalias, mibfile in self.get_mib_variants(mibname, **options):
                f = os.path.join(decode(path), decode(mibfile))

                debug.logger & debug.FLAG_READER and debug.logger(f"trying MIB {f}")

                if os.path.exists(f) and os.path.isfile(f):
                    try:
                        mtime = os.stat(f)[8]

                        debug.logger & debug.FLAG_READER and debug.logger(
                            f"source MIB {f} mtime is {time.strftime('%a, %d %b %Y %H:%M:%S GMT', time.gmtime(mtime))}, fetching data..."
                        )

                        fp = open(f, mode="rb")
                        mibData = fp.read(self.maxMibSize)
                        fp.close()

                        if len(mibData) == self.maxMibSize:
                            raise OSError(f"MIB {f} too large")

                        return MibInfo(
                            path=f"file://{f}",
                            file=mibfile,
                            name=mibalias,
                            mtime=mtime,
                        ), decode(mibData)

                    except OSError:
                        debug.logger & debug.FLAG_READER and debug.logger(
                            f"source file {f} open failure: {sys.exc_info()[1]}"
                        )

                        if not self._ignoreErrors:
                            raise error.PySmiError(
                                f"file {f} access error: {sys.exc_info()[1]}"
                            )

                    raise error.PySmiReaderFileNotModifiedError(
                        f"source MIB {f} is older than needed", reader=self
                    )

        raise error.PySmiReaderFileNotFoundError(
            f"source MIB {mibname} not found", reader=self
        )
